import asyncio
import functools
import subprocess
import time
from concurrent import futures
from typing import Any, Awaitable, Callable, List, Mapping, Optional, Union

import vcsinfo
from absl import logging

from lxm3._vendor.xmanager import xm
from lxm3._vendor.xmanager.xm import async_packager
from lxm3._vendor.xmanager.xm import id_predictor
from lxm3._vendor.xmanager.xm import job_blocks
from lxm3._vendor.xmanager.xm import pattern_matching as pm
from lxm3.xm_cluster import array_job as array_job_lib
from lxm3.xm_cluster import config as config_lib
from lxm3.xm_cluster import metadata
from lxm3.xm_cluster import packaging
from lxm3.xm_cluster.console import console
from lxm3.xm_cluster.execution import gridengine as gridengine_execution
from lxm3.xm_cluster.execution import local as local_execution
from lxm3.xm_cluster.execution import slurm as slurm_execution


class _LaunchResult:
    def __init__(self, local_handles, non_local_handles):
        self.local_handles = local_handles
        self.non_local_handles = non_local_handles


async def _launch(
    experiment_title: str,
    work_unit_name: str,
    job: Union[xm.JobGroup, array_job_lib.ArrayJob],
):
    local_handles = []
    non_local_handles = []

    job_name = f"{experiment_title}_{work_unit_name}"

    local_handles.extend(await local_execution.launch(job_name, job))  # type: ignore
    non_local_handles.extend(await slurm_execution.launch(job_name, job))  # type: ignore
    non_local_handles.extend(await gridengine_execution.launch(job_name, job))  # type: ignore

    return _LaunchResult(local_handles, non_local_handles)


class ClusterWorkUnit(xm.WorkUnit):
    """A mock version of WorkUnit with abstract methods implemented."""

    experiment: "ClusterExperiment"

    def __init__(
        self,
        experiment: "ClusterExperiment",
        work_unit_id_predictor: id_predictor.Predictor,
        create_task: Callable[[Awaitable[Any]], futures.Future[Any]],
        launched_jobs: List[job_blocks.JobType],
        launched_jobs_args: List[Optional[Mapping[str, Any]]],
        args: Optional[Mapping[str, Any]],
        role: xm.ExperimentUnitRole,
    ) -> None:
        super().__init__(experiment, create_task, args, role)
        self._launched_jobs = launched_jobs
        self._launched_jobs_args = launched_jobs_args
        self._work_unit_id = work_unit_id_predictor.reserve_id()
        self._work_unit_id_predictor = work_unit_id_predictor
        self._local_handles = []
        self._non_local_handles = []

    async def _launch_job_group(
        self,
        job_group: xm.JobGroup,
        args: Optional[Mapping[str, Any]],
        identity: str,
    ) -> None:
        """Appends the job group to the launched_jobs list."""
        del identity

        async with self._work_unit_id_predictor.submit_id(self._work_unit_id):  # type: ignore
            await self._submit_job_for_execution(job_group, args)

    async def _launch_job_config(self, job, args, identity):
        del identity
        assert not args
        async with self._work_unit_id_predictor.submit_id(self._work_unit_id):  # type: ignore
            await self._submit_job_for_execution(job, args)

    async def _submit_job_for_execution(
        self, job: Union[xm.JobGroup, array_job_lib.ArrayJob], args
    ):
        launch_result = await _launch(
            self.experiment._experiment_title, self.experiment_unit_name, job
        )
        self._ingest_handles(launch_result)

    def _ingest_handles(self, launch_result):
        """"""
        self._local_handles.extend(launch_result.local_handles)
        self._non_local_handles.extend(launch_result.non_local_handles)

    async def wait_for_local_jobs(self, is_exit_abrupt: bool):
        if not is_exit_abrupt:
            await asyncio.gather(*[handle.wait() for handle in self._local_handles])

    @property
    def work_unit_id(self) -> int:
        return self._work_unit_id

    @property
    def experiment_unit_name(self) -> str:
        return f"{self.experiment_id}_{self._work_unit_id}"

    @property
    def context(self) -> metadata.ClusterMetadataContext:
        return metadata.ClusterMetadataContext()


class ClusterExperiment(xm.Experiment):
    """A mock version of Experiment with abstract methods implemented."""

    _async_packager = async_packager.AsyncPackager(packaging.package)

    def __init__(
        self,
        experiment_title: str,
        vcs: Optional[vcsinfo.VCS] = None,
    ) -> None:
        super().__init__()
        self.launched_jobs = []
        self.launched_jobs_args = []
        self._work_units = []
        self._experiment_id = int(time.time() * 10**3)
        self._experiment_title = experiment_title
        self._vcs = vcs

    def _create_experiment_unit(
        self,
        args: Optional[Mapping[str, Any]],
        role: xm.ExperimentUnitRole = xm.WorkUnitRole(),
        identity: str = "",
    ) -> Awaitable[ClusterWorkUnit]:
        """Creates a new WorkUnit instance for the experiment."""
        del identity  # Unused.
        future = asyncio.Future(loop=self._event_loop)
        experiment_unit = ClusterWorkUnit(
            self,
            self._work_unit_id_predictor,
            self._create_task,
            self.launched_jobs,
            self.launched_jobs_args,
            args,
            role,
        )

        def _unsupported_aux_units(_):
            raise NotImplementedError("Auxiliary units are not supported")

        pm.match(
            pm.Case(
                [xm.WorkUnitRole],
                lambda _: self._work_units.append(experiment_unit),
            ),
            pm.Case(
                [xm.AuxiliaryUnitRole],
                _unsupported_aux_units,
            ),
        )(role)

        future.set_result(experiment_unit)
        return future

    def _wait_for_local_jobs(self, is_exit_abrupt: bool):
        if self._work_units:
            if any([wu._local_handles for wu in self._work_units]):
                console.print(
                    "Waiting for local jobs to complete. "
                    "Press Ctrl+C to terminate them and exit"
                )
        for unit in self._work_units:
            self._create_task(unit.wait_for_local_jobs(is_exit_abrupt))

    def __exit__(self, exc_type, exc_value, traceback):
        # Flush `.add` calls.
        self._wait_for_tasks()
        self._wait_for_local_jobs(exc_value is not None)
        return super().__exit__(exc_type, exc_value, traceback)

    async def __aexit__(self, exc_type, exc_value, traceback):
        # Flush `.add` calls.
        await self._await_for_tasks()
        self._wait_for_local_jobs(exc_value is not None)
        return await super().__aexit__(exc_type, exc_value, traceback)

    @property
    def work_unit_count(self) -> int:
        return len(self.work_units)

    @property
    def work_units(self):
        return self._work_units

    @property
    def experiment_id(self) -> int:
        return self._experiment_id

    @property
    def context(self) -> metadata.ClusterMetadataContext:
        return metadata.ClusterMetadataContext()


@functools.lru_cache()
def _load_vcsinfo() -> Optional[vcsinfo.VCS]:
    vcs = None

    try:
        vcs_root = subprocess.check_output(
            ["git", "rev-parse", "--show-toplevel"], text=True
        ).strip()
        vcs = vcsinfo.detect_vcs(vcs_root)
    except subprocess.SubprocessError:
        logging.debug("Failed to detect VCS info")

    return vcs


def create_experiment(experiment_title: str) -> ClusterExperiment:
    """Create a LXM3 experiment backed by the xm_cluster backend.
    Args:
        experiment_title: Title of the experiment.
        config: Optional config object to use. If set, override
            the configuration loaded from the config file.
    """
    config = config_lib.default()
    vcs = _load_vcsinfo()

    if not config.project() and vcs is not None:
        config.set_project(vcs.name)

    return ClusterExperiment(experiment_title, vcs=vcs)



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines convenience constants/functions for converting various units."""

# pylint: disable=invalid-name
vCPU = 1.0  # Virtual CPU

KiB = 2**10  # kibibyte
MiB = 2**20  # mibibyte
GiB = 2**30  # gibibyte
TiB = 2**40  # tebibyte
PiB = 2**50  # pebibyte

KB = 10**3  # kilobyte
MB = 10**6  # megabyte
GB = 10**9  # gigabyte
TB = 10**12  # terabyte
PB = 10**15  # petabyte

# Time units
Sec = 1
Min = 60
Hr = 60 * 60
# pylint: enable=invalid-name



import os
import tempfile
from typing import Any

from lxm3 import singularity
from lxm3 import xm
from lxm3._vendor.xmanager.xm import pattern_matching
from lxm3.xm_cluster import artifacts
from lxm3.xm_cluster import config as config_lib
from lxm3.xm_cluster import executable_specs as cluster_executable_specs
from lxm3.xm_cluster import executables as cluster_executables
from lxm3.xm_cluster.packaging import archive_builder
from lxm3.xm_cluster.packaging import container_builder


def _package_python_package(
    py_package: cluster_executable_specs.PythonPackage,
    packageable: xm.Packageable,
    artifact_store: artifacts.ArtifactStore,
):
    with tempfile.TemporaryDirectory() as staging:
        archive_name = archive_builder.create_python_archive(staging, py_package)
        local_archive_path = os.path.join(staging, archive_name)
        entrypoint_cmd = archive_builder.ENTRYPOINT_SCRIPT
        deployed_archive_path = artifact_store.deploy_resource_archive(
            local_archive_path
        )

    return cluster_executables.Command(
        entrypoint_command=entrypoint_cmd,
        resource_uri=deployed_archive_path,
        name=py_package.name,
        args=packageable.args,
        env_vars=packageable.env_vars,
    )


def _package_universal_package(
    universal_package: cluster_executable_specs.UniversalPackage,
    packageable: xm.Packageable,
    artifact_store: artifacts.ArtifactStore,
):
    with tempfile.TemporaryDirectory() as staging:
        archive_name = archive_builder.create_universal_archive(
            staging, universal_package
        )
        local_archive_path = os.path.join(staging, os.path.basename(archive_name))
        deployed_archive_path = artifact_store.deploy_resource_archive(
            local_archive_path
        )

    return cluster_executables.Command(
        entrypoint_command=" ".join(universal_package.entrypoint),
        resource_uri=deployed_archive_path,
        name=universal_package.name,
        args=packageable.args,
        env_vars=packageable.env_vars,
    )


def _package_pdm_project(
    pdm_project: cluster_executable_specs.PDMProject,
    packageable: xm.Packageable,
    artifact_store: artifacts.ArtifactStore,
):
    py_package = cluster_executable_specs.PythonPackage(
        pdm_project.entrypoint,
        path=pdm_project.path,
    )
    dockerfile = container_builder.pdm_dockerfile(
        pdm_project.base_image, pdm_project.lock_file
    )
    container_builder.build_docker_image(py_package.name, dockerfile, py_package.path)

    singularity_image = "docker-daemon://{}:latest".format(py_package.name)
    spec = cluster_executable_specs.SingularityContainer(py_package, singularity_image)
    return _package_singularity_container(spec, packageable, artifact_store)


def _package_python_container(
    python_container: cluster_executable_specs.PythonContainer,
    packageable: xm.Packageable,
    artifact_store: artifacts.ArtifactStore,
):
    py_package = cluster_executable_specs.PythonPackage(
        python_container.entrypoint, path=python_container.path
    )
    dockerfile = container_builder.python_container_dockerfile(
        base_image=python_container.base_image,
        requirements=python_container.requirements,
    )
    container_builder.build_docker_image(py_package.name, dockerfile, py_package.path)
    singularity_image = "docker-daemon://{}:latest".format(py_package.name)
    spec = cluster_executable_specs.SingularityContainer(py_package, singularity_image)
    return _package_singularity_container(spec, packageable, artifact_store)


def _package_singularity_container(
    container: cluster_executable_specs.SingularityContainer,
    packageable: xm.Packageable,
    artifact_store: artifacts.ArtifactStore,
):
    executable = _PACKAGING_ROUTER(container.entrypoint, packageable, artifact_store)

    singularity_image = container.image_path

    transport, _ = singularity.uri.split(singularity_image)
    if not transport:
        deploy_container_path = artifact_store.singularity_image_path(
            os.path.basename(singularity_image)
        )
        artifact_store.deploy_singularity_container(singularity_image)
    elif transport == "docker-daemon":
        # Try building singularity image using cache
        cache_image_path = container_builder.build_singularity_image_from_docker_daemon(
            singularity_image
        )
        deploy_container_path = artifact_store.singularity_image_path(
            os.path.basename(cache_image_path)
        )
        artifact_store.deploy_singularity_container(str(cache_image_path))
    else:
        deploy_container_path = singularity_image

    executable.singularity_image = deploy_container_path
    return executable


def _package_docker_container(
    container: cluster_executable_specs.DockerContainer,
    packageable: xm.Packageable,
    artifact_store: artifacts.ArtifactStore,
):
    executable = _PACKAGING_ROUTER(container.entrypoint, packageable, artifact_store)
    docker_image = container.image
    executable.docker_image = docker_image
    return executable


def _throw_on_unknown_executable(
    executable: Any,
    packageable: xm.Packageable,
    artifact_store: artifacts.ArtifactStore,
):
    del artifact_store
    raise TypeError(
        f"Unsupported executable specification: {executable!r}. "
        f"Packageable: {packageable!r}"
    )


_PACKAGING_ROUTER = pattern_matching.match(
    _package_python_package,
    _package_universal_package,
    _package_pdm_project,
    _package_python_container,
    _package_singularity_container,
    _package_docker_container,
    _throw_on_unknown_executable,
)


def package_for_cluster_executor(packageable: xm.Packageable):
    config = config_lib.default()
    cluster_settings = config.cluster_settings()

    artifact_store = artifacts.create_artifact_store(
        cluster_settings.storage_root,
        hostname=cluster_settings.hostname,
        user=cluster_settings.user,
        project=config.project(),
        connect_kwargs=cluster_settings.ssh_config,
    )
    return _PACKAGING_ROUTER(packageable.executable_spec, packageable, artifact_store)



import asyncio
import atexit
import concurrent.futures
import functools
import os
import shutil
import subprocess
from typing import List, Optional, Union

from absl import logging

from lxm3 import xm
from lxm3.xm_cluster import array_job as array_job_lib
from lxm3.xm_cluster import artifacts
from lxm3.xm_cluster import config as config_lib
from lxm3.xm_cluster import executables
from lxm3.xm_cluster import executors
from lxm3.xm_cluster.console import console
from lxm3.xm_cluster.execution import job_script

_LOCAL_EXECUTOR: Optional[concurrent.futures.ThreadPoolExecutor] = None


def local_executor():
    global _LOCAL_EXECUTOR
    if _LOCAL_EXECUTOR is None:
        _LOCAL_EXECUTOR = concurrent.futures.ThreadPoolExecutor(1)

        def shutdown():
            if _LOCAL_EXECUTOR is not None:
                logging.debug("Shutting down local executor...")
                _LOCAL_EXECUTOR.shutdown()

        atexit.register(shutdown)
    return _LOCAL_EXECUTOR


class LocalJobScriptBuilder(job_script.JobScriptBuilder):
    TASK_OFFSET = 1
    JOB_SCRIPT_SHEBANG = "#!/usr/bin/env bash"
    TASK_ID_VAR_NAME = "LOCAL_TASK_ID"

    @classmethod
    def _is_gpu_requested(cls, executor: executors.Local) -> bool:
        return shutil.which("nvidia-smi") is not None

    @classmethod
    def _create_setup_cmds(
        cls, executable: executables.Command, executor: executors.GridEngine
    ) -> str:
        del executor
        cmds = ["echo >&2 INFO[$(basename $0)]: Running on host $(hostname)"]

        if executable.singularity_image is not None:
            cmds.append(
                "echo >&2 INFO[$(basename $0)]: Singularity version: $(singularity --version)"
            )
        return "\n".join(cmds)

    @classmethod
    def _create_job_header(
        cls,
        executor: executors.GridEngine,
        num_array_tasks: Optional[int],
        job_script_dir: str,
        job_name: str,
    ) -> str:
        del executor, num_array_tasks, job_script_dir, job_name
        return ""

    def build(
        self,
        job: Union[xm.Job, array_job_lib.ArrayJob],
        job_name: str,
        job_script_dir: str,
    ) -> str:
        assert isinstance(job.executor, executors.Local)
        assert isinstance(job.executable, executables.Command)
        return super().build(job, job_name, job_script_dir)


class LocalExecutionHandle:
    def __init__(self, future: concurrent.futures.Future) -> None:
        self.future = future

    async def wait(self) -> None:
        return await asyncio.wrap_future(self.future)

    async def monitor(self) -> None:
        await asyncio.wrap_future(self.future)


def _local_job_predicate(job):
    if isinstance(job, xm.Job):
        return isinstance(job.executor, executors.Local)
    elif isinstance(job, array_job_lib.ArrayJob):
        return isinstance(job.executor, executors.Local)
    else:
        raise ValueError(f"Unexpected job type: {type(job)}")


class LocalClient(job_script.JobClient):
    builder_cls = LocalJobScriptBuilder

    def __init__(
        self,
        settings: Optional[config_lib.LocalSettings] = None,
        artifact_store: Optional[artifacts.LocalArtifactStore] = None,
    ) -> None:
        if settings is None:
            config = config_lib.default()
            settings = config.local_settings()
        self._settings = settings

        if artifact_store is None:
            config = config_lib.default()
            artifact_store = artifacts.LocalArtifactStore(
                self._settings.storage_root, project=config.project()
            )

        self._artifact_store = artifact_store

    def _launch(self, job_script_path, num_jobs):
        console.print(f"Launching {num_jobs} jobs locally...")
        handles = []
        for i in range(num_jobs):

            def task(i):
                subprocess.run(
                    ["bash", job_script_path],
                    env={
                        **os.environ,
                        LocalJobScriptBuilder.TASK_ID_VAR_NAME: str(
                            i + LocalJobScriptBuilder.TASK_OFFSET
                        ),
                    },
                )

            future = local_executor().submit(task, i)
            handles.append(LocalExecutionHandle(future))

        return None, handles


@functools.lru_cache()
def client() -> LocalClient:
    return LocalClient()


async def launch(job_name: str, job: job_script.ClusterJob):
    if isinstance(job, array_job_lib.ArrayJob):
        jobs = [job]  # type: ignore
    elif isinstance(job, xm.JobGroup):
        jobs: List[xm.Job] = xm.job_operators.flatten_jobs(job)
    elif isinstance(job, xm.Job):
        jobs = [job]

    jobs = [job for job in jobs if _local_job_predicate(job)]

    if not jobs:
        return []

    if len(jobs) > 1:
        raise ValueError(
            "Cannot launch a job group with multiple jobs as a single job."
        )

    return client().launch(job_name, jobs[0])



import copy
from typing import Dict, List, Optional

import attr

from lxm3 import xm


@attr.s(auto_attribs=True)
class ArrayJob(xm.JobConfig):
    executable: xm.Executable
    executor: xm.Executor
    name: Optional[str] = None
    args: List[xm.SequentialArgs] = attr.ib(
        converter=lambda args_list: list(
            map(xm.SequentialArgs.from_collection, args_list)
        ),
        default=attr.Factory(lambda: [xm.SequentialArgs()]),
    )
    env_vars: List[Dict[str, str]] = attr.ib(default=attr.Factory(lambda: [{}]))

    def __attrs_post_init__(self):
        self.args, self.env_vars = _broadcast_args(self.args, self.env_vars)


def _broadcast_args(args, env_vars):
    if len(args) == len(env_vars):
        return args, env_vars
    broadcast_size = max(len(args), len(env_vars))
    if min(len(args), len(env_vars)) != 1:
        raise ValueError(
            "args and env_vars must be the same length can be broadcast to same length"
        )
    if len(args) == 1:
        args = [copy.deepcopy(args[0]) for _ in range(broadcast_size)]
    if len(env_vars) == 1:
        env_vars = [copy.deepcopy(env_vars[0]) for _ in range(broadcast_size)]

    return args, env_vars


def flatten_array_job(job: ArrayJob) -> List[xm.Job]:
    jobs = []
    assert len(job.args) == len(job.env_vars)
    for args, env_vars in zip(job.args, job.env_vars):
        jobs.append(
            xm.Job(
                executable=job.executable,
                executor=job.executor,
                args=args,
                env_vars=env_vars,
            )
        )
    return jobs



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Data classes for job-related abstractions."""

import abc
import functools
import itertools
import re
from typing import Any, Awaitable, Callable, Dict, List, Mapping, Optional, Sequence, Tuple, TypeVar, Union

import attr
from lxm3._vendor.xmanager.xm import pattern_matching
from lxm3._vendor.xmanager.xm import utils

UserArgs = Union[Mapping, Sequence, 'SequentialArgs']


@functools.cache
def print_none_warning(key: str) -> None:
  print(
      f'WARNING: Setting `{key}=None` will exclude the flag. To pass the '
      f'actual value, pass the string literal `{key}="None"` instead'
  )


def _is_nested_structure(structure: Union[List[Any], Tuple[Any]]) -> bool:
  """Returns true if a list or tuple contains a list or tuple."""
  return any(type(element) in (list, tuple) for element in structure)


class SequentialArgs:
  """A sequence of positional and keyword arguments for a binary.

  Unix command line arguments are just a list of strings. But it is very common
  to simulate keyword arguments in a --key=value form. It is not uncommon to
  only have keyword arguments. Therefore we allow providing args as:

  Dicts:
    {'foo': 'space bar', 'with_magic': True} -> --foo='space bar' --with_magic
    Argument order is preserved.
  Lists:
    ['--foo', 'space bar'] -> --foo 'space bar'
  SequentialArgs (which allows to represent a mix of the two above):
    xm.merge_args({'foo': 'bar'}, ['--'], {'n': 16}) -> --foo=bar -- --n=16

  SequentialArgs provides a convenient merging semantics: if a value is given
  for an existing keyword argument, it will be overriden rather than appended,
  which allows to specify default values and override them later:

    xm.merge_args({'foo': '1', 'bar': '42'}, {'foo': '2'}) -> --foo=2 --bar=42

  SequentialArgs is immutable, but you can get a copy with updated value:

    args = xm.merge_args({'foo': '1', 'bar': '42'})
    args = xm.merge_args(args, {'foo': '2'})

  We only allow appending new arguments (positional and keyword) and overriding
  keyword arguments. Removal and inserting to the middle is not supported.
  """

  @attr.s(auto_attribs=True)
  class _RegularItem:
    value: Any

  @attr.s(auto_attribs=True)
  class _KeywordItem:
    name: str

  def __init__(self) -> None:
    """Constucts an empty SequentialArgs.

    Prefer using xm.merge_args to construct SequentialArgs objects.
    """
    self._items: List[
        Union[SequentialArgs._RegularItem, SequentialArgs._KeywordItem]
    ] = []
    self._kwvalues: Dict[str, Any] = {}

  def _ingest_regular_item(self, value: Any) -> None:
    self._items.append(SequentialArgs._RegularItem(value))

  def _ingest_keyword_item(self, name: str, value: Any) -> None:
    if name not in self._kwvalues:
      self._items.append(SequentialArgs._KeywordItem(name))
    self._kwvalues[name] = value

  def _merge_from(self, args: 'SequentialArgs') -> None:
    """Merges another instance of SequentialArgs into self."""

    def import_regular_item(item: SequentialArgs._RegularItem):
      self._ingest_regular_item(item.value)

    def import_keyword_item(item: SequentialArgs._KeywordItem):
      self._ingest_keyword_item(item.name, args._kwvalues[item.name])  # pylint: disable=protected-access

    matcher = pattern_matching.match(
        import_regular_item,
        import_keyword_item,
    )
    for item in args._items:  # pylint: disable=protected-access
      matcher(item)

  @staticmethod
  def from_collection(collection: Optional[UserArgs]) -> 'SequentialArgs':
    """Populates a new instance from a given collection."""
    result = SequentialArgs()
    if collection is None:
      return result

    def check_for_string(args: str) -> None:
      raise ValueError(
          f'Tried to construct xm.SequentialArgs from a string: {args!r}. '
          f'Wrap it in a list: [{args!r}] to make it a single argument.'
      )

    def import_sequential_args(args: SequentialArgs) -> None:
      result._merge_from(args)  # pylint: disable=protected-access

    def import_mapping(collection: Mapping[Any, Any]) -> None:
      for key, value in collection.items():
        result._ingest_keyword_item(str(key), value)  # pylint: disable=protected-access

    def import_sequence(collection: Sequence[Any]) -> None:
      for value in collection:
        result._ingest_regular_item(value)  # pylint: disable=protected-access

    matcher = pattern_matching.match(
        check_for_string,
        import_sequential_args,
        import_mapping,
        import_sequence,
    )
    matcher(collection)
    return result

  def rewrite_args(self, rewrite: Callable[[str], str]) -> 'SequentialArgs':
    """Applies the rewrite function to all args and returns the result."""
    result = SequentialArgs()

    def rewrite_regular_item(item: SequentialArgs._RegularItem):
      new_value = item.value
      if isinstance(new_value, str):
        new_value = rewrite(new_value)
      result._ingest_regular_item(new_value)  # pylint: disable=protected-access

    def rewrite_keyword_item(item: SequentialArgs._KeywordItem):
      new_value = self._kwvalues[item.name]
      if isinstance(new_value, str):
        new_value = rewrite(new_value)
      result._ingest_keyword_item(item.name, new_value)  # pylint: disable=protected-access

    matcher = pattern_matching.match(
        rewrite_regular_item,
        rewrite_keyword_item,
    )
    for item in self._items:
      matcher(item)

    return result

  def to_list(
      self,
      escaper: Callable[[Any], str] = utils.ARG_ESCAPER,
      kwargs_joiner: Callable[[str, str], str] = utils.trivial_kwargs_joiner,
  ) -> List[str]:
    """Exports items as a list ready to be passed into the command line."""

    def export_regular_item(
        item: SequentialArgs._RegularItem,
    ) -> List[Optional[str]]:
      return [escaper(item.value)]

    def export_keyword_item(
        item: SequentialArgs._KeywordItem,
    ) -> List[Optional[str]]:
      value = self._kwvalues[item.name]
      if value is None:
        # We skip flags with None value, allowing the binary to use defaults.
        # A string can be used if a literal "None" value needs to be assigned.
        print_none_warning(item.name)
        return [None]
      elif isinstance(value, bool):
        return [escaper(f"--{'' if value else 'no'}{item.name}")]
      elif type(value) in (list, tuple) and not _is_nested_structure(value):
        # Pass sequence of arguments in by repeating the flag for each
        # element to be consistent with absl's handling of multiple flags.
        # We do not do this for nested sequences, which absl cannot handle,
        # and instead fallback to quoting the sequence and leaving parsing of
        # the nested structure to the executable being called.
        return [
            kwargs_joiner(escaper(f'--{item.name}'), escaper(v)) for v in value
        ]
      else:
        return [kwargs_joiner(escaper(f'--{item.name}'), escaper(value))]

    matcher = pattern_matching.match(
        export_regular_item,
        export_keyword_item,
    )
    flags = itertools.chain.from_iterable(matcher(item) for item in self._items)
    return [f for f in flags if f is not None]

  def to_dict(self, kwargs_only: bool = False) -> Dict[str, Any]:
    """Exports items as a dictionary.

    Args:
      kwargs_only: Whether to skip positional arguments.

    Returns:
      The sought dictionary.
    """
    if kwargs_only:
      return self._kwvalues

    def export_regular_item(
        item: SequentialArgs._RegularItem,
    ) -> Tuple[str, Any]:
      return (str(item.value), True)

    def export_keyword_item(
        item: SequentialArgs._KeywordItem,
    ) -> Tuple[str, Any]:
      return (item.name, self._kwvalues[item.name])

    matcher = pattern_matching.match(
        export_regular_item,
        export_keyword_item,
    )
    return dict([matcher(item) for item in self._items])

  def __eq__(self, other) -> bool:
    return isinstance(other, SequentialArgs) and all([
        self._items == other._items,
        self._kwvalues == other._kwvalues,
    ])

  def __repr__(self) -> str:
    return f"[{', '.join(self.to_list(repr))}]"


def merge_args(*operands: Union[SequentialArgs, UserArgs]) -> SequentialArgs:
  """Merges several arguments collections into one left-to-right."""
  result = SequentialArgs()
  for operand in operands:
    if not isinstance(operand, SequentialArgs):
      operand = SequentialArgs.from_collection(operand)
    result._merge_from(operand)  # pylint: disable=protected-access
  return result


class ExecutableSpec(abc.ABC):
  """Executable specification describes what code / computation to run.

  Use one of the functions declared in xm/packagables.py to create a spec:
   * xm.binary - a prebuilt executable program.
   * xm.bazel_binary - an executable built with Bazel.
   * xm.container - a prebuilt Docker container.
   * xm.bazel_container - a Docker container built with Bazel.
   * xm.python_container - a Docker container running python code.
   * xm.dockerfile_container - a Docker container built with dockerfile.

  An executable spec must be turned into an Executable using
  Experiment.package() in order to be used in a Job.

  WARNING: `ExecutableSpec`s are supposed to be implementation-agnostic. That
  means there should be no backend-specific class inheriting `ExecutableSpec`.
  """

  @property
  @abc.abstractmethod
  def name(self) -> str:
    raise NotImplementedError


@attr.s(auto_attribs=True)
class Executable(abc.ABC):
  """Executable describes the final location of a packaged executable spec.

  An executable depends on the executable specification and the executor
  specification. Experiment's implementation knows how to handle each type of
  executable.

  Attributes:
    name: An automatically populated name for the executable. Used for assigning
      default names to `Job`s.
  """

  name: str


class ExecutorSpec(abc.ABC):
  """Executor spec describes the location of the runtime environment.

  For a list of supported ExecutorSpecs see a list of executors below.
  """


class Executor(abc.ABC):
  """Executor describes the runtime environment of a Job.

  Concrete supported executors are listed in xm_local/executors.py:
    * xm_local.Local
    * xm_local.Vertex
    * xm_local.Kubernetes
  """

  @classmethod
  @abc.abstractmethod
  def Spec(cls) -> ExecutorSpec:  # pylint: disable=invalid-name
    raise NotImplementedError


def _validate_env_vars(
    self: Any, attribute: Any, env_vars: Dict[str, str]
) -> None:
  del self  # Unused.
  del attribute  # Unused.
  for key in env_vars.keys():
    if not re.fullmatch('[a-zA-Z_][a-zA-Z0-9_]*', key):
      raise ValueError(
          'Environment variables names must conform to '
          f'[a-zA-Z_][a-zA-Z0-9_]*. Got {key!r}.'
      )


@attr.s(auto_attribs=True)
class Packageable:
  """Packageable describes what to build and its static parameters."""

  executable_spec: ExecutableSpec
  executor_spec: ExecutorSpec
  args: SequentialArgs = attr.ib(
      factory=list, converter=SequentialArgs.from_collection
  )  # pytype: disable=annotation-type-mismatch
  env_vars: Dict[str, str] = attr.ib(
      converter=dict, default=attr.Factory(dict), validator=_validate_env_vars
  )


class Constraint(abc.ABC):
  """Constraint describes the requirements for where a job group can run.

  Some examples of constraints include:

    * same virtual machine;
    * same virtual private Cloud subnetwork;
    * same network fabric;
    * same geographic location.
  """


# Job generators are async functions returning None.
# Pylint doesn't distinguish async and sync contexts so Optional[Awaitable] has
# to be used to accomodate both cases.
JobGeneratorType = Callable[..., Optional[Awaitable]]
JobType = Union['Job', 'JobGroup', JobGeneratorType, 'JobConfig']


@attr.s(auto_attribs=True)
class Job:
  """Job describes a unit of computation to be run.

  Attributes:
    executable: What to run -- one of `xm.Experiment.package` results.
    executor: Where to run -- one of `xm.Executor` subclasses.
    name: Name of the job. Must be unique within the context (work unit). By
      default it is constructed from the executable. Used for naming related
      entities such as newly created containers.
    args: Command line arguments to pass. This can be dict, list or
      xm.SequentialArgs. Dicts are most convenient for keyword flags.
      {'batch_size': 16} is passed as --batch_size=16. If positional arguments
      are needed one can use a list or xm.SequentialArgs.
    env_vars: Environment variables to apply.
  """

  executable: Executable
  executor: Executor
  name: Optional[str] = None
  args: SequentialArgs = attr.ib(
      factory=list, converter=SequentialArgs.from_collection
  )  # pytype: disable=annotation-type-mismatch
  env_vars: Dict[str, str] = attr.ib(
      converter=dict, default=attr.Factory(dict), validator=_validate_env_vars
  )


class JobGroup:
  """JobGroup describes a set of jobs that run under shared constraints.

  Use named arguments to give jobs meaningful names:

  ```
  JobGroup(
      learner=Job(learner_executable, executor),
      actor=Job(actor_executable, executor),
  )
  ```

  JobGroups provide the gang scheduling concept: Jobs inside them would be
  scheduled / descheduled simultaneously. Note that schedulers may not always be
  able to enforce that.

  JobGroups may include more fine grained constraints:

  ```
  JobGroup(
      learner=Job(tpu_learner_executable, executor),
      preprocessor=Job(preprocessor_executable, executor),
      constraints=[xm_impl.SameMachine()],
  )
  ```

  To express sophisticated requirements JobGroups can be nested:

  ```
  JobGroup(
      eval=Job(eval_executable, executor),
      colocated_learner_and_actor=JobGroup(
          learner=Job(tpu_learner_executable, executor),
          actor=Job(actor_executable, executor),
          constraints=[xm_impl.SameMachine()],
      ),
  )
  ```

  Attributes:
    jobs: A mapping of names to jobs.
    constraints: A list of additional scheduling constraints.
  """

  jobs: Dict[str, JobType]
  constraints: List[Constraint]

  def __init__(
      self,
      *,
      constraints: Optional[Sequence[Constraint]] = None,
      **jobs: JobType,
  ) -> None:
    """Builds a JobGroup.

    Args:
      constraints: List of additional scheduling constraints. Keyword only arg.
      **jobs: Jobs / job groups that constitute the group passed as kwargs.
    """
    self.jobs = jobs
    self.constraints = list(constraints) if constraints else []


class JobConfig(abc.ABC):
  """A job defined by a platform-specific configuration.

  Sometimes defining a job through a platform-agnostic xm.Job/xm.JobGroup
  interfaces is not feasible. In this case job can be defined by a configuration
  language native to the underlying platform. This is a base class for such
  configurations. Concrete XManager implementations may provide descendants for
  the configuration languages they support.
  """


JobTypeVar = TypeVar('JobTypeVar', Job, JobGroup, JobGeneratorType, JobConfig)


def get_args_for_all_jobs(job: JobType, args: Dict[str, Any]) -> Dict[str, Any]:
  """Gets args to apply on all jobs inside a JobGroup.

  This is useful if all jobs within a work unit accept the same arguments.

  Args:
    job: The job group to generate args for.
    args: The args to apply to all jobs inside the job group.

  Returns:
    args that can be added with work_unit.add()
  """
  if not isinstance(job, JobGroup):
    return {'args': dict(args)}
  all_args = {}
  for job_name, job_type in job.jobs.items():
    job_type_args = get_args_for_all_jobs(job_type, args)
    all_args[job_name] = job_type_args
  return all_args



from absl.testing import absltest
from absl.testing import parameterized

from lxm3 import xm
from lxm3.contrib import ucl
from lxm3.xm_cluster import requirements as cluster_requirements


class UCLClusterTest(parameterized.TestCase):
    @parameterized.named_parameters(
        (
            "gpu",
            cluster_requirements.JobRequirements(gpu=1, ram=8 * xm.GB),
            {"gpu": "true", "tmem": 8 * xm.GB},
            {},
        ),
        (
            "gpu_2",
            cluster_requirements.JobRequirements(gpu=2, ram=8 * xm.GB),
            {"gpu": "true", "tmem": 8 * xm.GB},
            {"gpu": 2},
        ),
        (
            "cpu",
            cluster_requirements.JobRequirements(ram=8 * xm.GB),
            {"tmem": 8 * xm.GB, "h_vmem": 8 * xm.GB},
            {},
        ),
        (
            "cpu_gpu_0",
            cluster_requirements.JobRequirements(gpu=0, ram=8 * xm.GB),
            {"tmem": 8 * xm.GB, "h_vmem": 8 * xm.GB},
            {},
        ),
        (
            "cpu_2",
            cluster_requirements.JobRequirements(gpu=0, cpu=2, ram=8 * xm.GB),
            {"tmem": 8 * xm.GB, "h_vmem": 8 * xm.GB},
            {"smp": 2},
        ),
    )
    def test_cs_cluster(self, requirements, expected_resources, expected_pe):
        executor = ucl._cs_executor_fn(requirements)
        self.assertEqual(executor.resources, expected_resources)
        self.assertEqual(executor.parallel_environments, expected_pe)

    @parameterized.named_parameters(
        (
            "gpu",
            cluster_requirements.JobRequirements(gpu=1, ram=8 * xm.GB),
            {"gpu": 1, "mem": 8 * xm.GB},
            {},
        ),
        (
            "gpu_2",
            cluster_requirements.JobRequirements(gpu=2, ram=8 * xm.GB),
            {"gpu": 2, "mem": 8 * xm.GB},
            {},
        ),
        (
            "cpu",
            cluster_requirements.JobRequirements(ram=8 * xm.GB),
            {"mem": 8 * xm.GB},
            {},
        ),
        (
            "cpu_gpu_0",
            cluster_requirements.JobRequirements(gpu=0, ram=8 * xm.GB),
            {"mem": 8 * xm.GB},
            {},
        ),
        (
            "cpu_2",
            cluster_requirements.JobRequirements(gpu=0, cpu=2, ram=8 * xm.GB),
            {"mem": 8 * xm.GB},
            {"smp": 2},
        ),
    )
    def test_myriad_cluster(self, requirements, expected_resources, expected_pe):
        executor = ucl._myriad_executor_fn(requirements)
        self.assertEqual(executor.resources, expected_resources)
        self.assertEqual(executor.parallel_environments, expected_pe)


if __name__ == "__main__":
    absltest.main()



import re
import shlex
import subprocess
import threading
from typing import Optional

import fabric


def parse_job_id(output: str) -> int:
    match = re.search(r"job (?P<id>[0-9]+)", output)
    if match is None:
        raise ValueError(f"Unable to parse job-id from:\n{output}")
    return int(match.group("id"))


class SlurmCluster:
    _connection: Optional[fabric.Connection] = None

    def __init__(
        self, hostname: Optional[str] = None, username: Optional[str] = None
    ) -> None:
        self._hostname = hostname
        self._username = username
        self._mutex = threading.Lock()
        self._connection = None
        if hostname is not None:
            self._connection = fabric.Connection(host=hostname, user=username)

    def close(self):
        with self._mutex:
            if self._connection is not None:
                self._connection.close()  # type: ignore

    def launch(self, command) -> int:
        output = self._submit_command(command)
        return parse_job_id(output)

    def _run_command(self, command: str) -> str:
        if self._connection is None:
            return subprocess.check_output(shlex.split(command), text=True)
        else:
            with self._mutex:
                return self._connection.run(command, hide="both").stdout

    def _submit_command(self, command: str) -> str:
        return self._run_command(f"sbatch {command}")

    def __repr__(self):
        if self._connection is not None:
            return f'Client(hostname="{self._hostname}", user="{self._username}")'
        else:
            return "Client()"



from absl import flags


def pytest_configure(config):
    del config  # Unused.
    flags.FLAGS.mark_as_parsed()


def pytest_ignore_collect(path, config):
    if path.isfile() and path.islink():
        return True



from lxm3._vendor.xmanager.xm import job_operators
from lxm3._vendor.xmanager.xm import executables
from lxm3._vendor.xmanager.xm import core
from lxm3._vendor.xmanager.xm import utils
from lxm3._vendor.xmanager.xm import job_blocks
from lxm3._vendor.xmanager.xm.compute_units import *  # noqa: F403
from lxm3._vendor.xmanager.xm.core import AuxiliaryUnitJob
from lxm3._vendor.xmanager.xm.core import AuxiliaryUnitRole
from lxm3._vendor.xmanager.xm.core import Experiment
from lxm3._vendor.xmanager.xm.core import ExperimentUnit
from lxm3._vendor.xmanager.xm.core import ExperimentUnitError
from lxm3._vendor.xmanager.xm.core import ExperimentUnitFailedError
from lxm3._vendor.xmanager.xm.core import ExperimentUnitNotCompletedError
from lxm3._vendor.xmanager.xm.core import ExperimentUnitRole
from lxm3._vendor.xmanager.xm.core import ExperimentUnitStatus
from lxm3._vendor.xmanager.xm.core import Importance
from lxm3._vendor.xmanager.xm.core import LaunchedJob
from lxm3._vendor.xmanager.xm.core import NotFoundError
from lxm3._vendor.xmanager.xm.core import WorkUnit
from lxm3._vendor.xmanager.xm.core import WorkUnitCompletedAwaitable
from lxm3._vendor.xmanager.xm.core import WorkUnitRole
from lxm3._vendor.xmanager.xm.job_blocks import Constraint
from lxm3._vendor.xmanager.xm.job_blocks import Executable
from lxm3._vendor.xmanager.xm.job_blocks import ExecutableSpec
from lxm3._vendor.xmanager.xm.job_blocks import Executor
from lxm3._vendor.xmanager.xm.job_blocks import ExecutorSpec
from lxm3._vendor.xmanager.xm.job_blocks import get_args_for_all_jobs
from lxm3._vendor.xmanager.xm.job_blocks import Job
from lxm3._vendor.xmanager.xm.job_blocks import JobConfig
from lxm3._vendor.xmanager.xm.job_blocks import JobGeneratorType
from lxm3._vendor.xmanager.xm.job_blocks import JobGroup
from lxm3._vendor.xmanager.xm.job_blocks import JobType
from lxm3._vendor.xmanager.xm.job_blocks import merge_args
from lxm3._vendor.xmanager.xm.job_blocks import Packageable
from lxm3._vendor.xmanager.xm.job_blocks import SequentialArgs
from lxm3._vendor.xmanager.xm.job_blocks import UserArgs
from lxm3._vendor.xmanager.xm.metadata_context import ContextAnnotations
from lxm3._vendor.xmanager.xm.metadata_context import MetadataContext
from lxm3._vendor.xmanager.xm.resources import GpuType
from lxm3._vendor.xmanager.xm.resources import InvalidTpuTopologyError
from lxm3._vendor.xmanager.xm.resources import JobRequirements
from lxm3._vendor.xmanager.xm.resources import ResourceDict
from lxm3._vendor.xmanager.xm.resources import ResourceQuantity
from lxm3._vendor.xmanager.xm.resources import ResourceType
from lxm3._vendor.xmanager.xm.resources import ServiceTier
from lxm3._vendor.xmanager.xm.resources import Topology
from lxm3._vendor.xmanager.xm.resources import TpuType
from lxm3._vendor.xmanager.xm.utils import run_in_asyncio_loop
from lxm3._vendor.xmanager.xm.utils import ShellSafeArg



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A mock implemenation of core XM objects.

This module is mainly purposed for unit tests. It provides implementation of
xm.Experiment, xm.WorkUnit, xm.MetadataContext which don't do much apart from
allowing to inspect what methods were called.
It also may be used as a stub implementation for cases when one doesn't care
how experiment is set up, but wants to maintain API consistency. For examle if
a framework provides own local run mode and wants to ignore any experiment
metadata manipulations.
"""

import asyncio
from concurrent import futures
from typing import Any, Awaitable, Callable, List, Mapping, Optional, Set

import attr
from lxm3._vendor.xmanager.xm import async_packager
from lxm3._vendor.xmanager.xm import core
from lxm3._vendor.xmanager.xm import id_predictor
from lxm3._vendor.xmanager.xm import job_blocks
from lxm3._vendor.xmanager.xm import metadata_context
from lxm3._vendor.xmanager.xm import pattern_matching as pm


class MockContextAnnotations(metadata_context.ContextAnnotations):
  """ContextAnnotations which stores all data in memory."""

  def __init__(self) -> None:
    self._title = ''
    self._tags = set()
    self._notes = ''

  @property
  def title(self) -> str:
    return self._title

  def set_title(self, title: str) -> None:
    self._title = title

  @property
  def tags(self) -> Set[str]:
    return self._tags

  def add_tags(self, *tags: str) -> None:
    self._tags.update(tags)

  def remove_tags(self, *tags: str) -> None:
    for tag in tags:
      self._tags.discard(tag)

  @property
  def notes(self) -> str:
    return self._notes

  def set_notes(self, notes: str) -> None:
    self._notes = notes


class MockMetadataContext(metadata_context.MetadataContext):
  """A MetadataContext which stores all data in memory."""

  def __init__(self) -> None:
    super().__init__(creator='unknown', annotations=MockContextAnnotations())


class MockExperimentUnit(core.WorkUnit):
  """A mock version of WorkUnit with abstract methods implemented."""

  def __init__(
      self,
      experiment: core.Experiment,
      work_unit_id_predictor: id_predictor.Predictor,
      create_task: Callable[[Awaitable[Any]], futures.Future[Any]],
      launched_jobs: List[job_blocks.JobType],
      launched_jobs_args: List[Optional[Mapping[str, Any]]],
      args: Optional[Mapping[str, Any]],
      role: core.ExperimentUnitRole,
  ) -> None:
    super().__init__(experiment, create_task, args, role)
    self._launched_jobs = launched_jobs
    self._launched_jobs_args = launched_jobs_args
    self._work_unit_id = work_unit_id_predictor.reserve_id()

  async def _wait_until_complete(self) -> None:
    """Mock work unit is immediately complete."""

  async def _launch_job_group(
      self,
      job_group: job_blocks.JobGroup,
      args: Optional[Mapping[str, Any]],
      identity: str,
  ) -> None:
    """Appends the job group to the launched_jobs list."""
    self._launched_jobs.extend(job_group.jobs.values())
    self._launched_jobs_args.append(args)

  @property
  def work_unit_id(self) -> int:
    return self._work_unit_id

  @property
  def experiment_unit_name(self) -> str:
    return f'{self.experiment_id}_{self._work_unit_id}'


class MockExperiment(core.Experiment):
  """A mock version of Experiment with abstract methods implemented."""

  constraints: List[job_blocks.JobType]

  _async_packager = async_packager.AsyncPackager(lambda _: [])

  def __init__(self) -> None:
    super().__init__()
    self.launched_jobs = []
    self.launched_jobs_args = []
    self._work_units = []
    self._auxiliary_units = []

    self._context = MockMetadataContext()

  def _create_experiment_unit(
      self,
      args: Optional[Mapping[str, Any]],
      role: core.ExperimentUnitRole = core.WorkUnitRole(),
      identity: str = '',
  ) -> Awaitable[MockExperimentUnit]:
    """Creates a new WorkUnit instance for the experiment."""
    del identity  # Unused.
    future = asyncio.Future(loop=self._event_loop)
    experiment_unit = MockExperimentUnit(
        self,
        self._work_unit_id_predictor,
        self._create_task,
        self.launched_jobs,
        self.launched_jobs_args,
        args,
        role,
    )
    pm.match(
        pm.Case(
            [core.WorkUnitRole],
            lambda _: self._work_units.append(experiment_unit),
        ),
        pm.Case(
            [core.AuxiliaryUnitRole],
            lambda _: self._auxiliary_units.append(experiment_unit),
        ),
    )(role)

    future.set_result(experiment_unit)
    return future

  @property
  def work_unit_count(self) -> int:
    return len(self.work_units)

  @property
  def work_units(self):
    return self._work_units

  @property
  def auxiliary_units(self):
    return self._auxiliary_units

  @property
  def experiment_id(self) -> int:
    return 1

  @property
  def context(self) -> MockMetadataContext:
    """Returns metadata context for the experiment."""
    return self._context


class MockExecutable(job_blocks.Executable):
  """A mock version of Executable with abstract methods implemented."""

  counter = 0

  def __init__(self):
    super().__init__(name=f'{MockExecutable.counter}')
    MockExecutable.counter += 1


class MockExecutor(job_blocks.Executor):
  """A mock version of Executor with abstract methods implemented."""

  Spec = job_blocks.ExecutorSpec  # pylint: disable=invalid-name


@attr.s(auto_attribs=True)
class MockConstraint(job_blocks.Constraint):
  id: str



import abc
import os
import re
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    TypedDict,
    TypeVar,
    Union,
    cast,
)

from lxm3 import xm
from lxm3._vendor.xmanager.xm import job_blocks
from lxm3.xm_cluster import array_job
from lxm3.xm_cluster import artifacts
from lxm3.xm_cluster import config as config_lib
from lxm3.xm_cluster import executables
from lxm3.xm_cluster import executors

ARRAY_WRAPPER_NAME = "array_wrapper.sh"
JOB_SCRIPT_NAME = "job.sh"

ClusterJob = Union[xm.Job, array_job.ArrayJob]

_JOB_SCRIPT_TEMPLATE = """\
%(shebang)s
%(header)s
%(setup)s
set -e
TASK_OFFSET=%(task_offset)s
TASK_INDEX_NAME="%(task_index_name)s"
NUM_TASKS="%(num_tasks)s"

if [ $NUM_TASKS -eq 1 ]; then
    # If there is only one task, then we don't need to use the task index
    LXM_TASK_ID=0
else
    LXM_TASK_ID=$(($(eval echo \\$"TASK_INDEX_NAME") - $TASK_OFFSET))
fi

export LXM_TASK_ID

LXM_WORKDIR=$(mktemp -d -t lxm-XXXXXX)
mkdir -p $LXM_WORKDIR
# Extract archives
ARCHIVES="%(archives)s"
for ar in $ARCHIVES; do
    case $ar in
    *.zip)
        unzip -q -d $LXM_WORKDIR $ar
        ;;
    *.tar)
        tar -C $LXM_WORKDIR -xf $ar
        ;;
    *.tar.gz|*.tgz)
        tar -C $LXM_WORKDIR -xzf $ar
        ;;
    *)
        _error "Unsupported archive format: $ar"
        ;;
    esac
done
cat <<'EOF' >$LXM_WORKDIR/job-env.sh
%(job_env)s
EOF
cat <<'EOF' >$LXM_WORKDIR/job-args.sh
%(job_args)s
EOF
chmod +x $LXM_WORKDIR/job-env.sh
chmod +x $LXM_WORKDIR/job-args.sh

cleanup() {
  echo >& 2 "DEBUG[$(basename $0)] Cleaning up $LXM_WORKDIR"
  rm -rf $LXM_WORKDIR
}
trap cleanup EXIT

cd $LXM_WORKDIR
%(entrypoint)s
"""


class _WorkItem(TypedDict):
    args: List[str]
    env_vars: Dict[str, str]


ExecutorClsType = TypeVar("ExecutorClsType")


def create_host_entrypoint(cmd: str) -> str:
    return """
CMD_ARGS=$(${LXM_WORKDIR}/job-args.sh)
sh -c ". $LXM_WORKDIR/job-env.sh; eval %(cmd)s ${CMD_ARGS}"
""" % {
        "cmd": cmd
    }


def _get_singularity_options(
    options: executors.SingularityOptions,
    use_gpu: bool,
) -> List[str]:
    opts = []
    if options.bind is not None:
        for host, container in options.bind.items():
            opts.extend(["--bind", f"{host}:{container}"])

    opts.extend(list(options.extra_options))

    if use_gpu:
        opts.append("--nv")

    return opts


def create_singularity_entrypoint(
    cmd: str,
    singularity_image: str,
    singularity_options,
    use_gpu: bool,
    extra_options=None,
) -> str:
    extra_options = extra_options or []
    options = singularity_options or executors.SingularityOptions()
    opts = _get_singularity_options(options, use_gpu)
    opts.extend(extra_options)

    return """\
singularity exec \
--bind="$LXM_WORKDIR/job-env.sh:/etc/job-env.sh:ro" \
--bind="$LXM_WORKDIR/job-args.sh:/etc/job-args.sh:ro" \
--bind="$LXM_WORKDIR:/mnt/lxm3" \
--pwd="/mnt/lxm3" \
--env="LXM_TASK_ID=$LXM_TASK_ID" \
%(singularity_opts)s \
%(singularity_image)s \
sh -c '. /etc/job-env.sh; eval %(cmd)s $(/etc/job-args.sh)'""" % {
        "singularity_image": singularity_image,
        "singularity_opts": " ".join(opts),
        "cmd": cmd,
    }


def _get_docker_options(
    options: executors.DockerOptions,
    use_gpu: bool,
) -> List[str]:
    opts = ["--rm"]
    if options.volumes is not None:
        for host, container in options.volumes.items():
            opts.extend(["--mount", f"type=bind,source={host},target={container}"])
    opts.extend(list(options.extra_options))

    if use_gpu:
        opts.extend(
            [
                "--runtime=nvidia",
                "-e",
                '"NVIDIA_VISIBLE_DEVICES=${CUDA_VISIBLE_DEVICES:-all}"',
            ]
        )

    return opts


def create_docker_entrypoint(
    cmd: str,
    docker_image: str,
    docker_options: executors.DockerOptions,
    use_gpu: bool,
) -> str:
    opts = _get_docker_options(docker_options, use_gpu)
    return """\
docker run \
--mount="type=bind,source=$LXM_WORKDIR/job-env.sh,target=/etc/job-env.sh,readonly" \
--mount="type=bind,source=$LXM_WORKDIR/job-args.sh,target=/etc/job-args.sh,readonly" \
--mount="type=bind,source=$LXM_WORKDIR,target=/mnt/lxm3" \
--workdir="/mnt/lxm3" \
--env="LXM_TASK_ID=$LXM_TASK_ID" \
%(docker_opts)s \
%(docker_image)s \
sh -c '. /etc/job-env.sh; eval %(cmd)s $(/etc/job-args.sh)'""" % {
        "docker_image": docker_image,
        "docker_opts": " ".join(opts),
        "cmd": cmd,
    }


def _get_additional_env(env_vars, parent_env):
    env = {}
    for k, v in parent_env.items():
        if k not in env_vars:
            env[k] = v
    return env


def _get_additional_binds(bind, singularity_additional_binds):
    updates = {}
    for src, dst in singularity_additional_binds.items():
        if dst not in bind.values():
            updates[src] = dst
    return updates


def _create_env_vars(env_vars_list: List[Dict[str, str]]) -> str:
    """Create the env_vars list."""
    lines = []
    first_keys = set(env_vars_list[0].keys())
    if not first_keys:
        return ""
    for env_vars in env_vars_list:
        if first_keys != set(env_vars.keys()):
            raise ValueError("Expect all environment variables to have the same keys")
    for key in first_keys:
        for task_id, env_vars in enumerate(env_vars_list):
            lines.append(
                '{key}_{task_id}="{value}"'.format(
                    key=key, task_id=task_id, value=env_vars[key]
                )
            )
        lines.append('{key}=$(eval echo \\$"{key}_$LXM_TASK_ID")'.format(key=key))
        lines.append("export {key}".format(key=key))
    content = "\n".join(lines)
    return content


def _create_args(args_list: List[List[str]]) -> str:
    """Create the args list."""
    if not args_list:
        return ":;"
    lines = []
    for task_id, args in enumerate(args_list):
        args_str = " ".join([a for a in args])
        lines.append(
            'TASK_CMD_ARGS_{task_id}="{args_str}"'.format(
                task_id=task_id, args_str=args_str
            )
        )
    lines.append('TASK_CMD_ARGS=$(eval echo \\$"TASK_CMD_ARGS_$LXM_TASK_ID")')
    lines.append("echo $TASK_CMD_ARGS")
    content = "\n".join(lines)
    return content


def _create_work_list(job: Union[xm.Job, array_job.ArrayJob]) -> List[_WorkItem]:
    work_list: List[_WorkItem] = []
    executable = cast(executables.Command, job.executable)
    if isinstance(job, array_job.ArrayJob):
        jobs = array_job.flatten_array_job(job)
    else:
        jobs = [job]
    for job in jobs:
        work_list.append(
            {
                "args": job_blocks.merge_args(executable.args, job.args).to_list(
                    xm.utils.ARG_ESCAPER
                ),
                "env_vars": {**executable.env_vars, **job.env_vars},
            }
        )
    return work_list


class JobScriptBuilder(abc.ABC):
    TASK_OFFSET: int
    JOB_SCRIPT_SHEBANG: str = "#!/usr/bin/env bash"
    TASK_ID_VAR_NAME: str

    def __init__(
        self,
        settings: Optional[config_lib.ExecutionSettings] = None,
    ) -> None:
        self._settings = settings

    @classmethod
    @abc.abstractmethod
    def _create_job_header(
        cls,
        executor: xm.Executor,
        num_array_tasks: Optional[int],
        job_script_dir: str,
        job_name: str,
    ) -> str:
        """Create a job header"""

    @classmethod
    @abc.abstractmethod
    def _is_gpu_requested(cls, executor: xm.Executor) -> bool:
        """Is GPU requested?"""

    @classmethod
    @abc.abstractmethod
    def _create_setup_cmds(cls, executable, executor) -> str:
        """Generate backend specific setup commands"""

    def _create_job_script(self, job: ClusterJob, setup: str, header: str) -> str:
        executable = job.executable
        if not isinstance(executable, executables.Command):
            raise ValueError("Only Command executable is supported")
        executor = cast(
            Union[executors.Local, executors.GridEngine, executors.Slurm], job.executor
        )
        use_gpu = self._is_gpu_requested(executor)

        if self._settings is not None:
            env_overrides = self._settings.env
            singularity_env_overrides = self._settings.env
            singularity_additional_binds = self._settings.singularity.binds
        else:
            env_overrides = {}
            singularity_env_overrides = {}
            singularity_additional_binds = {}

        singularity_image = executable.singularity_image
        singularity_options = executor.singularity_options
        docker_image = executable.docker_image
        docker_options = executor.docker_options
        if singularity_image is not None:
            if singularity_options is not None:
                singularity_options = executors.SingularityOptions(
                    bind={**singularity_options.bind},
                    extra_options=[*singularity_options.extra_options],
                )
            else:
                singularity_options = executors.SingularityOptions()

            singularity_options.bind.update(
                _get_additional_binds(
                    singularity_options.bind, singularity_additional_binds
                )
            )
        if docker_image is not None and docker_options is None:
            docker_options = executors.DockerOptions()

        if singularity_image is not None:
            assert singularity_options is not None
            entrypoint = create_singularity_entrypoint(
                executable.entrypoint_command,
                singularity_image,
                singularity_options,
                use_gpu,
            )
        elif docker_image is not None:
            assert docker_options is not None
            entrypoint = create_docker_entrypoint(
                executable.entrypoint_command, docker_image, docker_options, use_gpu
            )
        else:
            entrypoint = create_host_entrypoint(executable.entrypoint_command)

        work_list = _create_work_list(job)
        for work_item in work_list:
            if singularity_image is not None:
                work_item["env_vars"].update(
                    _get_additional_env(
                        work_item["env_vars"], singularity_env_overrides
                    )
                )
            work_item["env_vars"].update(
                _get_additional_env(work_item["env_vars"], env_overrides)
            )

        job_env = _create_env_vars([work["env_vars"] for work in work_list])
        job_args = _create_args([work["args"] for work in work_list])

        job_script_content = _JOB_SCRIPT_TEMPLATE % {
            "shebang": self.JOB_SCRIPT_SHEBANG,
            "header": header,
            "setup": setup,
            "job_env": job_env,
            "job_args": job_args,
            "task_offset": self.TASK_OFFSET,
            "entrypoint": entrypoint,
            "task_index_name": self.TASK_ID_VAR_NAME,
            "num_tasks": len(work_list),
            "archives": " ".join([executable.resource_uri]),
        }
        return job_script_content

    def build(
        self,
        job: Union[xm.Job, array_job.ArrayJob],
        job_name: str,
        job_script_dir: str,
    ) -> str:
        setup = self._create_setup_cmds(job.executable, job.executor)

        num_array_tasks = None
        if isinstance(job, array_job.ArrayJob) and len(job.args) > 1:
            num_array_tasks = len(job.args)
        header = self._create_job_header(
            job.executor, num_array_tasks, job_script_dir, job_name
        )

        return self._create_job_script(job=job, setup=setup, header=header)


class JobClient(abc.ABC):
    _artifact_store: artifacts.ArtifactStore
    _settings: config_lib.ExecutionSettings
    builder_cls: Callable[[config_lib.ExecutionSettings], JobScriptBuilder]

    @abc.abstractmethod
    def _launch(self, job_script_path: str, num_jobs: int) -> Tuple[Optional[str], Any]:
        raise NotImplementedError

    def launch(self, job_name: str, job: ClusterJob):
        job_name = re.sub("\\W", "_", job_name)
        job_script_dir = self._artifact_store.job_path(job_name)
        job_script_builder = self.builder_cls(self._settings)
        job_script_content = job_script_builder.build(job, job_name, job_script_dir)

        self._artifact_store.deploy_job_scripts(job_name, job_script_content)
        job_script_path = os.path.join(job_script_dir, JOB_SCRIPT_NAME)

        if isinstance(job, array_job.ArrayJob):
            num_jobs = len(job.env_vars)
        else:
            num_jobs = 1
        job_id, handles = self._launch(job_script_path, num_jobs)
        if job_id is not None:
            self._save_job_id(job_script_path, job_id)

        return handles

    def _save_job_id(self, job_script_path: str, job_id: str):
        self._artifact_store._fs.write_text(
            os.path.join(os.path.dirname(job_script_path), "job_id"), f"{job_id}\n"
        )



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for xmanager.xm.resources."""

from absl.testing import absltest
from absl.testing import parameterized

from lxm3._vendor.xmanager import xm
from lxm3._vendor.xmanager.xm import resources
from lxm3._vendor.xmanager.xm.resources import JobRequirements
from lxm3._vendor.xmanager.xm.resources import ResourceType


class ResourceDictTest(absltest.TestCase):

  def test_resource_type_by_name(self):
    self.assertEqual(ResourceType['cpu'], ResourceType.CPU)
    self.assertEqual(ResourceType['Cpu'], ResourceType.CPU)
    self.assertEqual(ResourceType['CPU'], ResourceType.CPU)
    with self.assertRaises(KeyError):
      ResourceType['UPC']  # pylint: disable=pointless-statement

  def test_resource_dict_to_string(self):
    resource_dict = resources.ResourceDict()
    resource_dict[ResourceType.V100] = 8
    resource_dict[ResourceType.CPU] = 4.2 * xm.vCPU
    resource_dict[ResourceType.MEMORY] = 16.5 * xm.GiB

    self.assertEqual(
        str(resource_dict), 'CPU: 4.2, MEMORY: 17716740096.0, V100: 8'
    )

  def test_resource_dict_from_job_requirements(self):
    requirements = JobRequirements(cpu=0.5 * xm.vCPU, memory=2 * xm.MiB, v100=8)
    resource_dict = requirements.task_requirements
    self.assertEqual(resource_dict[ResourceType.CPU], 0.5)
    self.assertEqual(resource_dict[ResourceType.MEMORY], 2097152)
    self.assertEqual(resource_dict[ResourceType.V100], 8)

  def test_job_requirements_unknown_key(self):
    with self.assertRaises(KeyError):
      JobRequirements(cpu=0.5 * xm.vCPU, upc=2)

  def test_requirements_summation(self):
    first = resources.JobRequirements(cpu=1, tpu_v2='2x2')
    second = resources.JobRequirements(ram=4 * xm.GiB, replicas=10)
    total = (
        first.replicas * first.task_requirements
        + second.replicas * second.task_requirements
    )
    self.assertEqual(total[ResourceType.CPU], 1)
    self.assertEqual(total[ResourceType.RAM], 40 * xm.GiB)
    self.assertEqual(total[ResourceType.TPU_V2], 4)


class TopologyTest(parameterized.TestCase):

  @parameterized.parameters(
      ('2', 2), ('4x4', 16), ('2x3x5', 30), ('4x4_twisted', 16)
  )
  def test_resource_type_by_name(self, topology, chip_count):
    self.assertEqual(resources.Topology(topology).chip_count, chip_count)

  def test_invalid_topology(self):
    with self.assertRaises(resources.InvalidTpuTopologyError):
      resources.Topology('euclidian')

  def test_topology_repr(self):
    self.assertEqual(repr(resources.Topology('4x4')), "xm.Topology('4x4')")

  def test_topology_eq(self):
    self.assertEqual(resources.Topology('4x4'), resources.Topology('4x4'))
    self.assertNotEqual(resources.Topology('2x2'), resources.Topology('4x4'))

    self.assertEqual(
        hash(resources.Topology('4x4')), hash(resources.Topology('4x4'))
    )


class JobRequirementsTest(parameterized.TestCase):

  def test_cpu_job(self):
    requirements = resources.JobRequirements(cpu=1.2, ram=1 * xm.GiB)
    self.assertEqual(
        requirements.task_requirements[resources.ResourceType.CPU], 1.2
    )
    self.assertEqual(
        requirements.task_requirements[resources.ResourceType.RAM], 1 * xm.GiB
    )
    self.assertIsNone(requirements.accelerator)
    self.assertIsNone(requirements.topology)
    self.assertEqual(requirements.replicas, 1)

  def test_construct_requirements(self):
    requirements = resources.JobRequirements(
        {resources.ResourceType.CPU: 4}, v100=1
    )
    task_requirements = requirements.task_requirements
    self.assertEqual(task_requirements[resources.ResourceType.CPU], 4)
    self.assertEqual(task_requirements[resources.ResourceType.V100], 1)
    self.assertEqual(requirements.replicas, 1)

  def test_resource_specified_twice(self):
    with self.assertRaises(ValueError):
      resources.JobRequirements({resources.ResourceType.CPU: 1}, cpu=2)

  def test_tpu_job(self):
    requirements = resources.JobRequirements(tpu_v3='4x4')
    self.assertEqual(requirements.accelerator, resources.ResourceType.TPU_V3)
    self.assertEqual(requirements.topology.name, '4x4')

  def test_multihost_gpu(self):
    requirements = resources.JobRequirements(v100='4x2')
    self.assertEqual(requirements.accelerator, resources.ResourceType.V100)
    self.assertEqual(requirements.topology.name, '4x2')
    self.assertEqual(requirements.replicas, 2)

  def test_multihost_gpu_with_cpu(self):
    requirements = resources.JobRequirements(v100='4x2', cpu=1)
    self.assertEqual(requirements.accelerator, resources.ResourceType.V100)
    self.assertEqual(requirements.topology.name, '4x2')
    self.assertEqual(requirements.replicas, 2)

  def test_multihost_gpu_with_replicas(self):
    with self.assertRaises(ValueError):
      resources.JobRequirements(v100='4x2', replicas=3)

  def test_location(self):
    requirements = resources.JobRequirements(location='lon_r7')
    self.assertEqual(requirements.location, 'lon_r7')

  def test_service_tier(self):
    requirements = resources.JobRequirements(
        service_tier=resources.ServiceTier.PROD
    )
    self.assertEqual(requirements.service_tier, resources.ServiceTier.PROD)

  def test_service_tier_mutable(self):
    requirements = resources.JobRequirements(
        service_tier=resources.ServiceTier.PROD
    )
    requirements.service_tier = resources.ServiceTier.BATCH
    self.assertEqual(requirements.service_tier, resources.ServiceTier.BATCH)

  def test_replicas(self):
    requirements = resources.JobRequirements(replicas=2)
    self.assertEqual(requirements.replicas, 2)

    with self.assertRaises(ValueError):
      resources.JobRequirements(replicas=2, tpu_v3='4x4')

    resources.JobRequirements(replicas=2, v100='4x2')
    resources.JobRequirements(v100='4x2')
    with self.assertRaises(ValueError):
      resources.JobRequirements(replicas=4, v100='4x2')

  def test_str(self):
    self.assertEqual(
        repr(
            resources.JobRequirements(
                cpu=1,
                location='lon_r7',
                service_tier=resources.ServiceTier.BATCH,
                replicas=2,
            )
        ),
        (
            "xm.JobRequirements(cpu=1.0, location='lon_r7',"
            ' service_tier=xm.ServiceTier.BATCH, replicas=2)'
        ),
    )

  def test_str_omits_empty_fields(self):
    self.assertEqual(
        repr(resources.JobRequirements(cpu=1)), 'xm.JobRequirements(cpu=1.0)'
    )

  def test_is_gpu_tpu_given_cpu(self):
    requirements = resources.JobRequirements(cpu=1, ram=4 * xm.GiB)
    self.assertNotIn(requirements.accelerator, xm.GpuType)
    self.assertNotIn(requirements.accelerator, xm.TpuType)

  def test_is_gpu_tpu_given_gpu(self):
    requirements = resources.JobRequirements(cpu=1, v100=4)
    self.assertIn(requirements.accelerator, xm.GpuType)
    self.assertNotIn(requirements.accelerator, xm.TpuType)

  def test_is_gpu_tpu_given_tpu(self):
    requirements = resources.JobRequirements(cpu=1, tpu_v2='2x2')
    self.assertNotIn(requirements.accelerator, xm.GpuType)
    self.assertIn(requirements.accelerator, xm.TpuType)


class EnumSubsetTest(parameterized.TestCase):

  def test_construction(self):
    self.assertEqual(resources.GpuType['V100'], resources.GpuType(17))
    with self.assertRaises(AttributeError):
      resources.GpuType['TPU_V3']  # pylint: disable=pointless-statement
    with self.assertRaises(ValueError):
      resources.GpuType(170)

  def test_equivalence(self):
    self.assertEqual(resources.ResourceType.V100, resources.GpuType.V100)

  def test_iteration(self):
    gpus = set(iter(resources.GpuType))
    self.assertIn(resources.ResourceType.V100, gpus)
    self.assertNotIn(resources.ResourceType.TPU_V3, gpus)

  def test_contains(self):
    self.assertIn(resources.ResourceType.V100, resources.GpuType)
    self.assertNotIn(resources.ResourceType.TPU_V3, resources.GpuType)


if __name__ == '__main__':
  absltest.main()



import datetime
from typing import Any, Dict, Optional, Sequence, Union

import attr

from lxm3 import xm
from lxm3.xm_cluster.requirements import JobRequirements


def _convert_time(
    time: Optional[Union[str, datetime.datetime, int]]
) -> Optional[datetime.timedelta]:
    if time is None:
        return None
    if isinstance(time, int):
        return datetime.timedelta(seconds=time)
    elif isinstance(time, datetime.timedelta):
        return time
    else:
        raise TypeError(
            f"Expect walltime to be type int (seconds) or datetime.timedelta, got {type(time)}"
        )


@attr.s(auto_attribs=True)
class SingularityOptions:
    """Options for singularity container.

    Args:
        bind: Mapping of the form ``{src: dst}``.
        User-bind path specification of the form ``-B <src>:<dst>``.
        extra_options: Extra commandline options to pass to singularity.

          When using the ``extra_options``, be aware of the following:
            1. lxm3 currently probably won't likely work with
               ``--contain``, ``--pwd`` and ``-wd``,
            2. You don't have to pass ``--nv`` for GPU jobs,
               lxm3 will do it for you.
    """

    bind: Dict[str, str] = attr.Factory(dict)
    extra_options: Sequence[str] = attr.Factory(list)


@attr.s(auto_attribs=True)
class DockerOptions:
    """Options for singularity container.

    Args:
        bind: Mapping of the form ``{src: dst}``.
        User-bind path specification of the form ``-B <src>:<dst>``.
        extra_options: Extra commandline options to pass to singularity.

          When using the ``extra_options``, be aware of the following:
            1. lxm3 currently probably won't likely work with
               ``--contain``, ``--pwd`` and ``-wd``,
            2. You don't have to pass ``--nv`` for GPU jobs,
               lxm3 will do it for you.
    """

    volumes: Dict[str, str] = attr.Factory(dict)
    extra_options: Sequence[str] = attr.Factory(list)


@attr.s(auto_attribs=True)
class LocalSpec(xm.ExecutorSpec):
    """Spec for local execution."""


@attr.s(auto_attribs=True)
class Local(xm.Executor):
    """Local executor.

    Args:
        requirements: placeholder, no effect right now
        singularity_options: Options for singularity container
    """

    requirements: JobRequirements = attr.Factory(JobRequirements)

    singularity_options: Optional[SingularityOptions] = None
    docker_options: Optional[DockerOptions] = None

    @classmethod
    def Spec(cls) -> LocalSpec:
        return LocalSpec()


@attr.s(auto_attribs=True)
class GridEngineSpec(xm.ExecutorSpec):
    """Spec for SGE execution."""


@attr.s(auto_attribs=True)
class GridEngine(xm.Executor):
    """SGE executor.

    Attributes:
        requirements: placeholder, no effect right now.
        resources: Resources passed to qsub as `-l key=value`.
        parallel_environments: Parallel environments in the form of ``--pe <name> <slots>``.
        walltime: Maximum running time, ``-l h_rt=time``.
            When an ``int`` is used, this is interpreted as seconds.
            A ``datetime.timedelta`` can also be used.
        queue: queue to submit the job to: ``-q``.
        reserved: If set, use ``-R y``.
        log_directory: Log directory for stdout/stderr.
        merge_output: If False, log to separate files.
        shell: Shell to use, default ``/bin/bash``.
        project: ``-P``.
        account: ``-A``.
        modules: Modules to load before running the job.
            See https://modules.readthedocs.io/en/latest/
        max_parallel_tasks: ``-tc``.
        extra_directives: Extra directives to pass to ``qsub``.
        skip_directives: Directives to skip.
        singularity_options: Options for singularity container.

    """

    # WARNING:
    # requirements are currently ignored as different SGE clusters
    # use different approaches for configurting these resources.
    # To configure, use resources for -l directives.
    # For UCL clusters, use the auto-configuration package from lxm3.contrib.
    requirements: JobRequirements = attr.Factory(JobRequirements)
    # Resources passed to qsub as -l
    resources: Dict[str, Any] = attr.Factory(dict)
    # Parallel environments in the form of --pe <name> <slots>
    parallel_environments: Dict[str, int] = attr.Factory(dict)
    # Maximum running time, -l h_rt
    walltime: Optional[datetime.timedelta] = attr.field(
        default=None, converter=_convert_time
    )

    # queue to submit the job to: -q
    queue: Optional[str] = None
    # If set, use -R y
    reserved: Optional[bool] = None
    # Log directory for stdout/stderr
    log_directory: Optional[str] = None
    # If False, log to separate files
    merge_output: bool = True
    shell: str = "/bin/bash"

    # -P
    project: Optional[str] = None
    # -A
    account: Optional[str] = None

    # Modules to load before running the job
    modules: Sequence[str] = attr.Factory(list)

    # -tc
    max_parallel_tasks: Optional[int] = None
    extra_directives: Sequence[str] = attr.Factory(list)
    skip_directives: Sequence[str] = attr.Factory(list)

    singularity_options: Optional[SingularityOptions] = None
    docker_options: Optional[DockerOptions] = None

    @classmethod
    def Spec(cls) -> GridEngineSpec:
        return GridEngineSpec()


@attr.s(auto_attribs=True)
class SlurmSpec(xm.ExecutorSpec):
    """Spec for Slurm execution."""


@attr.s(auto_attribs=True)
class Slurm(xm.Executor):
    """Slurm executor."""

    requirements: JobRequirements = attr.Factory(JobRequirements)
    resources: Dict[str, Any] = attr.Factory(dict)
    walltime: Optional[datetime.timedelta] = attr.field(
        default=None, converter=_convert_time
    )

    singularity_options: Optional[SingularityOptions] = None
    docker_options: Optional[DockerOptions] = None

    log_directory: Optional[str] = None
    # Modules to load before running the job
    modules: Sequence[str] = attr.Factory(list)

    exclusive: bool = False
    partition: Optional[str] = None

    extra_directives: Sequence[str] = attr.Factory(list)
    skip_directives: Sequence[str] = attr.Factory(list)

    @classmethod
    def Spec(cls) -> SlurmSpec:
        return SlurmSpec()



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""XManager client API.

Provides XManager public API for configuring and launching experiments.
"""

from lxm3._vendor.xmanager.xm import job_operators
from lxm3._vendor.xmanager.xm.compute_units import *
from lxm3._vendor.xmanager.xm.core import AuxiliaryUnitJob
from lxm3._vendor.xmanager.xm.core import AuxiliaryUnitRole
from lxm3._vendor.xmanager.xm.core import Experiment
from lxm3._vendor.xmanager.xm.core import ExperimentUnit
from lxm3._vendor.xmanager.xm.core import ExperimentUnitError
from lxm3._vendor.xmanager.xm.core import ExperimentUnitFailedError
from lxm3._vendor.xmanager.xm.core import ExperimentUnitNotCompletedError
from lxm3._vendor.xmanager.xm.core import ExperimentUnitRole
from lxm3._vendor.xmanager.xm.core import ExperimentUnitStatus
from lxm3._vendor.xmanager.xm.core import Importance
from lxm3._vendor.xmanager.xm.core import LaunchedJob
from lxm3._vendor.xmanager.xm.core import NotFoundError
from lxm3._vendor.xmanager.xm.core import WorkUnit
from lxm3._vendor.xmanager.xm.core import WorkUnitCompletedAwaitable
from lxm3._vendor.xmanager.xm.core import WorkUnitRole
from lxm3._vendor.xmanager.xm.executables import BazelBinary
from lxm3._vendor.xmanager.xm.executables import BazelContainer
from lxm3._vendor.xmanager.xm.executables import Binary
from lxm3._vendor.xmanager.xm.executables import BinaryDependency
from lxm3._vendor.xmanager.xm.executables import CommandList
from lxm3._vendor.xmanager.xm.executables import Container
from lxm3._vendor.xmanager.xm.executables import Dockerfile
from lxm3._vendor.xmanager.xm.executables import ModuleName
from lxm3._vendor.xmanager.xm.executables import PythonContainer
from lxm3._vendor.xmanager.xm.job_blocks import Constraint
from lxm3._vendor.xmanager.xm.job_blocks import Executable
from lxm3._vendor.xmanager.xm.job_blocks import ExecutableSpec
from lxm3._vendor.xmanager.xm.job_blocks import Executor
from lxm3._vendor.xmanager.xm.job_blocks import ExecutorSpec
from lxm3._vendor.xmanager.xm.job_blocks import get_args_for_all_jobs
from lxm3._vendor.xmanager.xm.job_blocks import Job
from lxm3._vendor.xmanager.xm.job_blocks import JobConfig
from lxm3._vendor.xmanager.xm.job_blocks import JobGeneratorType
from lxm3._vendor.xmanager.xm.job_blocks import JobGroup
from lxm3._vendor.xmanager.xm.job_blocks import JobType
from lxm3._vendor.xmanager.xm.job_blocks import merge_args
from lxm3._vendor.xmanager.xm.job_blocks import Packageable
from lxm3._vendor.xmanager.xm.job_blocks import SequentialArgs
from lxm3._vendor.xmanager.xm.job_blocks import UserArgs
from lxm3._vendor.xmanager.xm.metadata_context import ContextAnnotations
from lxm3._vendor.xmanager.xm.metadata_context import MetadataContext
from lxm3._vendor.xmanager.xm.packagables import bazel_binary
from lxm3._vendor.xmanager.xm.packagables import bazel_container
from lxm3._vendor.xmanager.xm.packagables import binary
from lxm3._vendor.xmanager.xm.packagables import container
from lxm3._vendor.xmanager.xm.packagables import dockerfile_container
from lxm3._vendor.xmanager.xm.packagables import python_container
from lxm3._vendor.xmanager.xm.resources import GpuType
from lxm3._vendor.xmanager.xm.resources import InvalidTpuTopologyError
from lxm3._vendor.xmanager.xm.resources import JobRequirements
from lxm3._vendor.xmanager.xm.resources import ResourceDict
from lxm3._vendor.xmanager.xm.resources import ResourceQuantity
from lxm3._vendor.xmanager.xm.resources import ResourceType
from lxm3._vendor.xmanager.xm.resources import ServiceTier
from lxm3._vendor.xmanager.xm.resources import Topology
from lxm3._vendor.xmanager.xm.resources import TpuType
from lxm3._vendor.xmanager.xm.utils import run_in_asyncio_loop
from lxm3._vendor.xmanager.xm.utils import ShellSafeArg



import functools
import os
from typing import Any, Dict, Optional, Protocol

import appdirs
import tomlkit
from absl import flags
from absl import logging

LXM_CONFIG = flags.DEFINE_string(
    "lxm_config", os.environ.get("LXM_CONFIG"), "Path to lxm configuration."
)


class SingularitySettings:
    def __init__(self, data) -> None:
        self._data = data

    def __repr__(self) -> str:
        return repr(self._data)

    @property
    def binds(self) -> Dict[str, str]:
        binds = self._data.get("binds", [])
        return {bind["src"]: bind["dest"] for bind in binds}

    @property
    def env(self) -> Dict[str, str]:
        return self._data.get("env", {})


class ExecutionSettings(Protocol):
    @property
    def env(self) -> Dict[str, str]:
        ...

    @property
    def singularity(self) -> SingularitySettings:
        ...


class LocalSettings(ExecutionSettings):
    def __init__(self, data=None) -> None:
        self._data = data or {}

    def __repr__(self) -> str:
        return repr(self._data)

    @property
    def storage_root(self) -> str:
        return self._data["storage"]["staging"]

    @property
    def env(self) -> Dict[str, str]:
        return self._data.get("env", {})

    @property
    def singularity(self) -> SingularitySettings:
        return SingularitySettings(self._data.get("singularity", {}))


class ClusterSettings(ExecutionSettings):
    def __init__(self, data=None) -> None:
        self._data = data or {}

    def __repr__(self) -> str:
        return repr(self._data)

    @property
    def storage_root(self):
        return self._data["storage"]["staging"]

    @property
    def hostname(self):
        return self._data.get("server", None)

    @property
    def user(self):
        return self._data.get("user", None)

    @property
    def ssh_config(self):
        connect_kwargs = {}
        proxycommand = self._data.get("proxycommand", None)
        if proxycommand is not None:
            import paramiko

            connect_kwargs["sock"] = paramiko.ProxyCommand(proxycommand)

        ssh_private_key = self._data.get("ssh_private_key", None)
        if ssh_private_key is not None:
            connect_kwargs["key_filename"] = os.path.expanduser(ssh_private_key)

        password = self._data.get("password", None)
        if password is not None:
            connect_kwargs["password"] = password

        return connect_kwargs

    @property
    def env(self) -> Dict[str, str]:
        return self._data.get("env", {})

    @property
    def singularity(self) -> SingularitySettings:
        return SingularitySettings(self._data.get("singularity", {}))


class Config:
    def __init__(self, data) -> None:
        self._data = data

    def __repr__(self) -> Any:
        return repr(self._data)

    @classmethod
    def from_file(cls, path: str) -> "Config":
        with open(path, "rt") as f:
            config_dict = tomlkit.loads(f.read())
        return cls(config_dict)

    @classmethod
    def from_string(cls, content: str) -> "Config":
        config_dict = tomlkit.loads(content)
        return cls(config_dict)

    def project(self) -> Optional[str]:
        project = os.environ.get("LXM_PROJECT", None)
        if project is not None:
            return project
        return self._data.get("project", None)

    def set_project(self, project):
        self._data["project"] = project

    def local_settings(self) -> LocalSettings:
        return LocalSettings(self._data["local"])

    def default_cluster(self) -> str:
        cluster = os.environ.get("LXM_CLUSTER", None)
        if cluster is None:
            cluster = self._data["clusters"][0]["name"]
        return cluster

    def cluster_settings(self) -> ClusterSettings:
        location = self.default_cluster()
        clusters = {cluster["name"]: cluster for cluster in self._data["clusters"]}
        if location not in clusters:
            raise ValueError("Unknown cluster")
        cluster_config = clusters[location]
        return ClusterSettings(cluster_config)


@functools.lru_cache()
def default() -> Config:
    # The path may be relative, especially if it comes from sys.argv[0].
    try:
        if LXM_CONFIG.value is not None:
            logging.debug("Loading config from %s", LXM_CONFIG.value)
            return Config.from_file(LXM_CONFIG.value)
    except flags.UnparsedFlagAccessError:
        logging.debug("Unable to load config from flag")
    # Use file from environment variable
    lxm_config_path = os.environ.get("LXM_CONFIG", None)
    if lxm_config_path is not None:
        logging.debug("Loading config from %s", lxm_config_path)
        return Config.from_file(lxm_config_path)
    cwd_path = os.path.join(os.getcwd(), "lxm.toml")
    if os.path.exists(cwd_path):
        logging.debug("Loading config from %s", cwd_path)
        return Config.from_file(cwd_path)
    user_config_path = os.path.join(appdirs.user_config_dir("lxm3"), "config.toml")
    if os.path.exists(user_config_path):
        logging.debug("Loading config from %s", cwd_path)
        return Config.from_file(user_config_path)
    else:
        raise ValueError("Unable to load Config")



#!/usr/bin/env python3
import os

from absl import app
from absl import flags

from lxm3 import xm
from lxm3 import xm_cluster

_SINGULARITY_IMAGE = flags.DEFINE_string(
    "singularity_image", None, "Name of singularity image"
)
_DOCKER_IMAGE = flags.DEFINE_string("docker_image", None, "Name of docker image")


def main(_):
    with xm_cluster.create_experiment(experiment_title="basic") as experiment:
        job_requirements = xm_cluster.JobRequirements(ram=8 * xm.GB)
        executor = xm_cluster.Local(
            job_requirements,
            singularity_options=xm_cluster.SingularityOptions(),
            docker_options=xm_cluster.DockerOptions(
                extra_options=[f"--user={os.getuid()}:{os.getgid()}"]
            ),
        )

        spec = xm_cluster.PythonPackage(
            # This is a relative path to the launcher that contains
            # your python package (i.e. the directory that contains pyproject.toml)
            path=".",
            # Entrypoint is the python module that you would like to
            # In the implementation, this is translated to
            #   python3 -m py_package.main
            entrypoint=xm_cluster.ModuleName("py_package.main"),
        )

        if _SINGULARITY_IMAGE.value is not None:
            spec = xm_cluster.SingularityContainer(
                spec,
                image_path=_SINGULARITY_IMAGE.value,
            )
        elif _DOCKER_IMAGE.value is not None:
            spec = xm_cluster.DockerContainer(
                spec,
                image=_DOCKER_IMAGE.value,
            )

        [executable] = experiment.package(
            [xm.Packageable(spec, executor_spec=executor.Spec())]
        )

        experiment.add(
            xm.Job(executable=executable, executor=executor, args={"seed": 1})
        )


if __name__ == "__main__":
    app.run(main)



import os
import pathlib
import subprocess
import tempfile

import appdirs

from lxm3 import singularity
from lxm3.xm_cluster.console import console


def pdm_dockerfile(base_image: str, lock_file: str):
    return """\
FROM {base_image} as builder
RUN if ! id 1000; then useradd -m -u 1000 docker; fi
RUN pip install pdm

COPY {lock_file} /app/pdm.lock
COPY pyproject.toml /app/pyproject.toml
COPY README.md /app/README.md

WORKDIR /app
RUN pdm install && pdm export > /requirements.txt

FROM {base_image}
COPY --from=builder /requirements.txt /requirements.txt
RUN pip install --no-cache-dir -r /requirements.txt
""".format(
        base_image=base_image, lock_file=lock_file
    )


def python_container_dockerfile(base_image: str, requirements: str):
    return """\
FROM {base_image}
RUN if ! id 1000; then useradd -m -u 1000 docker; fi

COPY {requirements} /requirements.txt
RUN pip install --no-cache-dir -r /requirements.txt
""".format(
        base_image=base_image, requirements=requirements
    )


def build_docker_image(image_name: str, dockerfile_content: str, path: str):
    with tempfile.TemporaryDirectory() as tmpdir:
        dockerfile = os.path.join(tmpdir, "Dockerfile")
        with open(dockerfile, "w") as f:
            f.write(dockerfile_content)
        subprocess.run(
            ["docker", "buildx", "build", "-t", image_name, "-f", dockerfile, path],
            check=True,
        )


def build_singularity_image_from_docker_daemon(singularity_image: str) -> str:
    transport, ref = singularity.uri.split(singularity_image)
    if transport != "docker-daemon":
        raise ValueError(
            f"Expected docker-daemon transport, got {transport} for {singularity_image}"
        )

    filename = singularity.uri.filename(singularity_image, "sif")
    build_cache_dir = pathlib.Path(appdirs.user_cache_dir("lxm3"), "singularity")
    cache_image_path = build_cache_dir / filename
    marker_file = cache_image_path.with_suffix(cache_image_path.suffix + ".image_id")
    try:
        import docker
    except ImportError:
        raise ValueError(
            "docker-py library is required to build Singularity images from docker-daemon"
        )

    client = docker.from_env()
    image_id: str = client.images.get(ref).id  # type: ignore

    should_rebuild = True
    if cache_image_path.exists():
        if marker_file.exists():
            old_image_id = pathlib.Path(marker_file).read_text().strip()
            should_rebuild = old_image_id != image_id
            if should_rebuild:
                console.log("Image ID changed, rebuilding...")
            else:
                console.log("Reusing cached image from", cache_image_path)
        else:
            should_rebuild = True
            console.log("Marker file does not exist, rebuilding...")
    else:
        should_rebuild = True
        console.log("Container cache does not exist, rebuilding...")

    if should_rebuild:
        cache_image_path.parent.mkdir(parents=True, exist_ok=True)
        singularity.images.build_singularity_image(
            cache_image_path, singularity_image, force=True
        )
        console.log("Cached image at", cache_image_path)
        pathlib.Path(marker_file).write_text(image_id)

    return str(cache_image_path)



import functools
import os
from typing import List, Optional

from absl import logging

from lxm3 import xm
from lxm3.clusters import gridengine
from lxm3.xm_cluster import array_job
from lxm3.xm_cluster import artifacts
from lxm3.xm_cluster import config as config_lib
from lxm3.xm_cluster import executables
from lxm3.xm_cluster import executors
from lxm3.xm_cluster import requirements as cluster_requirements
from lxm3.xm_cluster.console import console
from lxm3.xm_cluster.execution import job_script


class GridEngineJobScriptBuilder(job_script.JobScriptBuilder):
    TASK_OFFSET = 1
    JOB_SCRIPT_SHEBANG = "#!/usr/bin/env bash"
    TASK_ID_VAR_NAME = "SGE_TASK_ID"

    executable_cls = executables.Command
    executor_cls = executors.GridEngine

    @classmethod
    def _is_gpu_requested(cls, executor: executors.GridEngine) -> bool:
        return (
            "gpu" in executor.parallel_environments
            or cluster_requirements.ResourceType.GPU
            in executor.requirements.task_requirements
            or "gpu" in executor.resources
        )

    @classmethod
    def _create_setup_cmds(
        cls, executable: executables.Command, executor: executors.GridEngine
    ) -> str:
        cmds = ["echo >&2 INFO[$(basename $0)]: Running on host $(hostname)"]

        for module in executor.modules:
            cmds.append(f"module load {module}")
        if cls._is_gpu_requested(executor):
            cmds.append(
                "echo >&2 INFO[$(basename $0)]: CUDA_VISIBLE_DEVICES=$CUDA_VISIBLE_DEVICES"
            )
            cmds.append("nvidia-smi")

        if executable.singularity_image is not None:
            cmds.append(
                "echo >&2 INFO[$(basename $0)]: Singularity version: $(singularity --version)"
            )
        return "\n".join(cmds)

    @classmethod
    def _create_job_header(
        cls,
        executor: executors.GridEngine,
        num_array_tasks: Optional[int],
        job_script_dir: str,
        job_name: str,
    ) -> str:
        job_header = header_from_executor(
            job_name, executor, num_array_tasks, job_script_dir
        )
        return job_header

    def build(
        self, job: job_script.ClusterJob, job_name: str, job_script_dir: str
    ) -> str:
        assert isinstance(job.executor, self.executor_cls)
        assert isinstance(job.executable, self.executable_cls)
        return super().build(job, job_name, job_script_dir)


class GridEngineHandle:
    def __init__(self, job_id: str) -> None:
        self.job_id = job_id

    async def wait(self):
        raise NotImplementedError()

    async def monitor(self):
        raise NotImplementedError()


def _sge_job_predicate(job):
    if isinstance(job, xm.Job):
        return isinstance(job.executor, executors.GridEngine)
    elif isinstance(job, array_job.ArrayJob):
        return isinstance(job.executor, executors.GridEngine)
    else:
        raise ValueError(f"Unexpected job type: {type(job)}")


class GridEngineClient(job_script.JobClient):
    builder_cls = GridEngineJobScriptBuilder
    _settings: config_lib.ClusterSettings

    def __init__(
        self,
        settings: Optional[config_lib.ClusterSettings] = None,
        artifact_store: Optional[artifacts.ArtifactStore] = None,
    ) -> None:
        if settings is None:
            settings = config_lib.default().cluster_settings()
        self._settings = settings

        if artifact_store is None:
            project = config_lib.default().project()
            artifact_store = artifacts.create_artifact_store(
                settings.storage_root,
                hostname=settings.hostname,
                user=settings.user,
                project=project,
                connect_kwargs=settings.ssh_config,
            )
        self._artifact_store = artifact_store

        self._cluster = gridengine.GridEngineCluster(
            self._settings.hostname, self._settings.user
        )

    def _launch(self, job_script_path, num_jobs):
        cluster = self._cluster
        console.log(f"Launching {num_jobs} job on {self._settings.hostname}")

        console.log(f"Launch with command:\n  qsub {job_script_path}")
        group = cluster.launch(job_script_path)
        console.log(f"Successfully launched job {group.group(0)}")
        job_ids = gridengine.split_job_ids(group)

        return group.group(0), [GridEngineHandle(job_id) for job_id in job_ids]


@functools.lru_cache()
def client():
    return GridEngineClient()


async def launch(job_name: str, job: job_script.ClusterJob) -> List[GridEngineHandle]:
    if isinstance(job, array_job.ArrayJob):
        jobs = [job]  # type: ignore
    elif isinstance(job, xm.JobGroup):
        jobs: List[xm.Job] = xm.job_operators.flatten_jobs(job)
    else:
        raise NotImplementedError()

    jobs = [job for job in jobs if _sge_job_predicate(job)]

    if not jobs:
        return []

    if len(jobs) > 1:
        raise ValueError(
            "Cannot launch a job group with multiple jobs as a single job."
        )

    return client().launch(job_name, jobs[0])


def _format_time(duration_seconds: int) -> str:
    hours, remainder = divmod(duration_seconds, 3600)
    minutes, seconds = divmod(remainder, 60)
    return "{:02}:{:02}:{:02}".format(int(hours), int(minutes), int(seconds))


def header_from_executor(
    job_name: str,
    executor: executors.GridEngine,
    num_array_tasks: Optional[int],
    job_script_dir: str,
) -> str:
    header = []

    header.append(f"#$ -N {job_name}")

    if executor.requirements.task_requirements:
        logging.warning(
            "Task requirements in GridEngine will be ignored."
            "Please use GridEngine.resources for customizing resources."
        )

    for resource, value in executor.resources.items():
        if value:
            header.append(f"#$ -l {resource}={value}")

    for pe_name, value in executor.parallel_environments.items():
        if value > 0:
            header.append(f"#$ -pe {pe_name} {value}")

    if executor.walltime is not None:
        duration_secs = int(executor.walltime.total_seconds())
        header.append(f"#$ -l h_rt={_format_time(duration_secs)}")

    if executor.queue:
        header.append(f"#$ -q {executor.queue}")

    reserved = executor.reserved
    if reserved is None:
        if (
            executor.parallel_environments
            or cluster_requirements.ResourceType.GPU
            in executor.requirements.task_requirements
            or "gpu" in executor.resources
        ):
            reserved = True

    if reserved:
        header.append("#$ -R y")

    log_directory = executor.log_directory or os.path.join(job_script_dir, "logs")
    if num_array_tasks is not None:
        stdout = os.path.join(log_directory, "$JOB_NAME.o$JOB_ID.$TASK_ID")
        stderr = os.path.join(log_directory, "$JOB_NAME.e$JOB_ID.$TASK_ID")
    else:
        stdout = os.path.join(log_directory, "$JOB_NAME.o$JOB_ID")
        stderr = os.path.join(log_directory, "$JOB_NAME.e$JOB_ID")

    header.append(f"#$ -o {stdout}")
    header.append(f"#$ -e {stderr}")
    if executor.merge_output:
        header.append("#$ -j y")

    if num_array_tasks is not None:
        header.append(f"#$ -t 1-{num_array_tasks}")
        if executor.max_parallel_tasks:
            header.append(f"#$ -tc {executor.max_parallel_tasks}")

    # Use current working directory
    header.append("#$ -cwd")

    header.append(f"#$ -S {executor.shell}")

    if executor.project:
        header.append(f"#$ -P {executor.project}")

    if executor.account:
        header.append(f"#$ -A {executor.account}")

    # Skip requested header directives
    header = list(
        filter(
            lambda line: not any(skip in line for skip in executor.skip_directives),
            header,
        )
    )

    for line in executor.extra_directives:
        if not line.startswith("#$"):
            line = "#$ " + line
        header.append(line)

    return "\n".join(header)



from unittest import mock

from absl.testing import absltest
from absl.testing import parameterized

from lxm3 import xm
from lxm3 import xm_cluster
from lxm3.xm_cluster import array_job
from lxm3.xm_cluster import config as config_lib
from lxm3.xm_cluster import experiment as cluster_experiment

_TEST_CONFIG = """
project = ""
[local]
[local.storage]
staging = "{}"

[[clusters]]
name = "default"
server = "localhost"
user = "user"

[clusters.storage]
staging = "/home/foo/lxm3-staging"
"""


class DummyHandle:
    async def wait(self):
        return


def _fake_launch(job):
    print(job)
    if isinstance(job, xm_cluster.ArrayJob):
        jobs = array_job.flatten_array_job(job)
    else:
        jobs = xm.job_operators.flatten_jobs(job)

    return cluster_experiment._LaunchResult(
        local_handles=[DummyHandle()] * len(jobs), non_local_handles=[]
    )


class ExperimentTest(parameterized.TestCase):
    def setUp(self):
        super().setUp()
        staging = self.create_tempdir().full_path
        self._config = config_lib.Config.from_string(_TEST_CONFIG.format(staging))
        self._executable = xm_cluster.Command(
            name="dummy",
            entrypoint_command="./entrypint.sh",
            resource_uri="dummy",
        )

    @mock.patch.object(cluster_experiment, "_launch")
    def test_create_experiment(self, mock_launch):
        with mock.patch.object(config_lib, "default") as mock_default:
            mock_default.return_value = self._config
            job = xm.Job(executable=self._executable, executor=xm_cluster.Local())
            jobs = [job] * 2
            mock_launch.return_value = _fake_launch(job)
            experiment = xm_cluster.create_experiment("test")
            with experiment:
                for job in jobs:
                    experiment.add(job)
            self.assertEqual(experiment.work_unit_count, 2)
            self.assertEqual(mock_launch.call_count, 2)

    @mock.patch.object(cluster_experiment, "_launch")
    def test_create_experiment_batch(self, mock_launch):
        with mock.patch.object(config_lib, "default") as mock_default:
            mock_default.return_value = self._config
            job = xm_cluster.ArrayJob(
                executable=self._executable,
                executor=xm_cluster.Local(),
                args=[{"seed": 1}, {"seed": 2}],
            )
            mock_launch.return_value = _fake_launch(job)
            experiment = xm_cluster.create_experiment("test")
            with experiment:
                experiment.add(job)
            self.assertEqual(experiment.work_unit_count, 1)
            mock_launch.assert_called_once()


if __name__ == "__main__":
    absltest.main()



# Port of https://github.com/sylabs/singularity/blob/main/internal/pkg/util/uri/uri.go
from typing import Optional

# Library is the keyword for a library ref
Library = "library"
# Shub is the keyword for a shub ref
Shub = "shub"
# HTTP is the keyword for http ref
HTTP = "http"
# HTTPS is the keyword for https ref
HTTPS = "https"
# Oras is the keyword for an oras ref
Oras = "oras"

_VALID_URIS: dict[str, bool] = {
    "library": True,
    "shub": True,
    "docker": True,
    "docker-archive": True,
    "docker-daemon": True,
    "oci": True,
    "oci-archive": True,
    "http": True,
    "https": True,
    "oras": True,
}


# is_valid returns whether or not the given source is valid
def is_valid(source: str) -> tuple[bool, Optional[str]]:
    u = source.split(":", 2)

    if len(u) != 2:
        return False, f"invalid uri {source}"

    if u[0] in _VALID_URIS:
        return True, ""

    return False, f"invalid uri {source}"


def filename(uri: str, suffix: str) -> str:
    """
    filename turns a transport:ref URI into a filename containing the top-level
    identifier of the image. For example, docker://sylabsio/lolcow:latest returns
    lolcow_latest.<suffix>

    Returns "" when not in transport:ref format
    """
    transport, ref = split(uri)
    if transport == "":
        return ""

    ref = ref.lstrip("/")  # Trim leading "/" characters
    refSplit = ref.split("/")  # Split ref into parts

    if transport == HTTP or transport == HTTPS:
        imageName = refSplit[len(refSplit) - 1]
        return imageName

    # Default tag is latest
    tags = ["latest"]
    container = refSplit[len(refSplit) - 1]

    if ":" in container:
        imageParts = container.split(":")
        container = imageParts[0]
        tags = [imageParts[1]]
        if "," in tags[0]:
            tags = tags[0].split(",")

    return "{}_{}.{}".format(container, tags[0], suffix)


def split(uri: str) -> tuple[str, str]:
    """
    Split splits a URI into it's components which can be used directly through containers/image
    This can be tricky if there is no type but a file name contains a colon.

    Examples:
        docker://ubuntu -> docker, //ubuntu
        docker://ubuntu:18.04 -> docker, //ubuntu:18.04
        oci-archive:path/to/archive -> oci-archive, path/to/archive
        ubuntu -> "", ubuntu
        ubuntu:18.04.img -> "", ubuntu:18.04.img

    Returns:
        transport: str, ref: str
    """
    uriSplit = uri.split(":", 1)
    if len(uriSplit) == 1:
        # no colon
        return "", uri

    if uriSplit[1].startswith("//"):
        # the format was ://, so try it whether or not valid URI
        return uriSplit[0], uriSplit[1]

    ok, err = is_valid(uri)
    if ok and not err:
        # also accept recognized URIs
        return uriSplit[0], uriSplit[1]

    return "", uri



import datetime
import os
from typing import Optional

import fsspec
from rich.prompt import Confirm

from lxm3.xm_cluster import config as config_lib
from lxm3.xm_cluster.console import console

_JOB = "job"
_ARCHIVE = "archive"
_CONTAINER = "container"
_VALID_TYPES = [_JOB, _ARCHIVE, _CONTAINER]


class ClusterJob:
    def __init__(self, client, job_root) -> None:
        self._client = client
        self._job_root = job_root

    @property
    def job_name(self):
        return os.path.basename(self._job_root)

    def job_script(self):
        return (
            self._client._fs.cat(os.path.join(self._job_root, "job.sh"))
            .decode("utf-8")
            .strip()
        )

    def job_id(self):
        return (
            self._client._fs.cat(os.path.join(self._job_root, "job_id"))
            .decode("utf-8")
            .strip()
        )

    def logs(self):
        logs = self._client._fs.ls(os.path.join(self._job_root, "logs"))
        return logs

    def time_created(self):
        return self._client._fs.info(os.path.join(self._job_root, "job_id"))["mtime"]


class JobClient:
    def __init__(self, cluster_settings) -> None:
        self._cluster_settings = cluster_settings
        self._fs = fsspec.filesystem(
            "sftp",
            host=cluster_settings.hostname,
            username=cluster_settings.user,
            **cluster_settings.ssh_config,
        )

    @property
    def filesystem(self):
        return self._fs

    def list_projects(self):
        project_root = os.path.join(self._cluster_settings.storage_root, "projects")
        return [os.path.basename(p) for p in self._fs.ls(project_root)]

    def list_jobs(self, project):
        job_root = os.path.join(
            self._cluster_settings.storage_root, "projects", project, "jobs"
        )
        return [os.path.basename(p) for p in self._fs.ls(job_root)]

    def list_containers(self, project):
        container_root = os.path.join(
            self._cluster_settings.storage_root, "projects", project, "containers"
        )
        return [p for p in self._fs.ls(container_root)]

    def list_archives(self, project):
        archive_root = os.path.join(
            os.path.join(
                self._cluster_settings.storage_root, "projects", project, "archives"
            )
        )
        return [p for p in self._fs.ls(archive_root)]

    def get_job(self, project, job_name):
        job_path = os.path.join(
            self._cluster_settings.storage_root, "projects", project, "jobs", job_name
        )

        return ClusterJob(self, job_path)


def run_clean(
    project: str,
    days: Optional[int],
    dry_run: bool = False,
    force: bool = False,
    type_: Optional[str] = None,
):
    config = config_lib.default()
    client = JobClient(config.cluster_settings())
    now = datetime.datetime.now(tz=datetime.timezone.utc)

    if type_ is None:
        types = _VALID_TYPES
    else:
        types = type_.split(",")
        for type_ in types:
            if type_ not in _VALID_TYPES:
                raise ValueError("Invalid type: {}".format(type_))

    valid_projects = client.list_projects()
    if project not in valid_projects:
        raise ValueError(
            "Invalid project: {}, available projects are {}".format(
                project, valid_projects
            )
        )
    items_to_remove = []
    if _JOB in types:
        for job_name in client.list_jobs(project):
            job = client.get_job(project, job_name)
            job_root = job._job_root
            time_created = client.filesystem.info(job_root)["mtime"]
            if not days or (now - time_created > datetime.timedelta(days=days)):
                items_to_remove.append(("job", job_root))

    if _ARCHIVE in types:
        for archive in client.list_archives(project):
            time_created = client.filesystem.info(archive)["mtime"]
            if not days or (now - time_created > datetime.timedelta(days=days)):
                items_to_remove.append(("archive", archive))

    if _CONTAINER in types:
        for container in client.list_containers(project):
            time_created = client.filesystem.info(container)["mtime"]
            if not days or (now - time_created > datetime.timedelta(days=days)):
                items_to_remove.append(("container", container))

    if dry_run:
        if len(items_to_remove) > 0:
            for item_type, path in items_to_remove:
                console.print("Removing {} {}".format(item_type, path))
        else:
            console.print("No items to remove")
            return
    else:
        remove = False
        if not force:
            if len(items_to_remove) == 0:
                console.print("No items to remove")
                return
            for item_type, path in items_to_remove:
                console.print("Would remove [bold]{}[bold] {}".format(item_type, path))
            try:
                remove = Confirm.ask("Do you wish to continue?")
            except KeyboardInterrupt:
                console.print("Aborting")
                return
        else:
            remove = True

        if remove:
            for item_type, path in items_to_remove:
                console.print("Removing [bold]{}[bold] {}".format(item_type, path))
                try:
                    client.filesystem.rm(path, recursive=True)
                except:  # noqa
                    console.print_exception()
                    console.print("Failed to remove {}".format(path))



from typing import Sequence

from lxm3 import xm
from lxm3._vendor.xmanager.xm import pattern_matching
from lxm3.xm_cluster import executors
from lxm3.xm_cluster.packaging import cluster as cluster_packaging
from lxm3.xm_cluster.packaging import local as local_packaging


def packaging_router(
    packageable: xm.Packageable,
):
    def package_local(executable_spec: executors.LocalSpec):
        return local_packaging.package_for_local_executor(packageable)

    def package_gridengine(executable_spec: executors.GridEngineSpec):
        return cluster_packaging.package_for_cluster_executor(packageable)

    def package_slurm(executable_spec: executors.SlurmSpec):
        return cluster_packaging.package_for_cluster_executor(packageable)

    return pattern_matching.match(
        package_local,
        package_gridengine,
        package_slurm,
    )(packageable.executor_spec)


def package(packageables: Sequence[xm.Packageable]):
    executables = []

    for packageable in packageables:
        executables.append(packaging_router(packageable))

    return executables



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from lxm3._vendor.xmanager import xm_mock
from lxm3._vendor.xmanager.xm import job_blocks
from lxm3._vendor.xmanager.xm import job_operators


def construct_job(name=None):
  return job_blocks.Job(
      name=name,
      executable=xm_mock.MockExecutable(),
      executor=xm_mock.MockExecutor(),
  )


class JobOperatorsTest(unittest.TestCase):

  def test_collect_jobs_by_filter_gathers_matches(self):
    job_group = job_blocks.JobGroup(
        foo=construct_job('foo'),
        bar=construct_job('bar'),
        baz=construct_job('baz'),
    )

    self.assertEqual(
        job_operators.collect_jobs_by_filter(
            job_group,
            predicate=lambda job: job.name in ['foo', 'baz'],
        ),
        [job_group.jobs['foo'], job_group.jobs['baz']],
    )

  def test_flatten_jobs_traverses_nested_groups(self):
    baz = construct_job('baz')
    foo = construct_job('foo')
    job_group = job_blocks.JobGroup(
        foo=foo,
        bar=job_blocks.JobGroup(baz=baz),
    )

    self.assertEqual(
        job_operators.flatten_jobs(job_group),
        [foo, baz],
    )

  def test_aggregate_constraint_cliques(self):
    outer_1 = construct_job('outer_1')
    inner_1 = construct_job('inner_1')
    inner_2 = construct_job('inner_2')
    constraint_a = xm_mock.MockConstraint('A')
    constraint_b = xm_mock.MockConstraint('B')
    constraint_c = xm_mock.MockConstraint('C')
    job_group = job_blocks.JobGroup(
        outer_1=outer_1,
        outer_2=job_blocks.JobGroup(
            inner_1=inner_1,
            inner_2=inner_2,
            constraints=[constraint_b, constraint_c],
        ),
        constraints=[constraint_a],
    )

    self.assertEqual(
        job_operators.aggregate_constraint_cliques(job_group),
        [
            job_operators.ConstraintClique(
                constraint=constraint_a,
                jobs=[outer_1, inner_1, inner_2],
            ),
            job_operators.ConstraintClique(
                constraint=constraint_b,
                jobs=[inner_1, inner_2],
            ),
            job_operators.ConstraintClique(
                constraint=constraint_c,
                jobs=[inner_1, inner_2],
            ),
        ],
    )


if __name__ == '__main__':
  unittest.main()



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from unittest import mock

from absl.testing import absltest
from lxm3._vendor.xmanager.xm import job_blocks


class JobBlocksTest(unittest.TestCase):

  def test_from_mapping(self):
    args = job_blocks.SequentialArgs.from_collection({
        'a': 1,
        'b': 2,
        'c': 3,
    })

    self.assertEqual(args.to_list(str), ['--a=1', '--b=2', '--c=3'])

  def test_from_mapping_multi(self):
    args = job_blocks.SequentialArgs.from_collection({
        'a': 1,
        'c': [3, '4'],
    })

    self.assertEqual(args.to_list(str), ['--a=1', '--c=3', '--c=4'])

  def test_from_mapping_nested_multi(self):
    args = job_blocks.SequentialArgs.from_collection({'a': [[1, 2, 3]]})

    self.assertEqual(args.to_list(str), ['--a=[[1, 2, 3]]'])

  def test_from_sequence(self):
    args = job_blocks.SequentialArgs.from_collection([1, 2, 3])

    self.assertEqual(args.to_list(str), ['1', '2', '3'])

  def test_from_none(self):
    args = job_blocks.SequentialArgs.from_collection(None)
    self.assertEqual(args.to_list(str), [])

  def test_merge_args(self):
    args = job_blocks.merge_args(
        [1],
        {
            'a': 'z',
            'b': 'y',
        },
        [2],
        {
            'b': 'x',
            'c': 't',
        },
        [3],
    )

    self.assertEqual(
        args.to_list(str), ['1', '--a=z', '--b=x', '2', '--c=t', '3']
    )

  def test_to_dict(self):
    args = job_blocks.merge_args(['--knob'], {1: False})

    self.assertEqual(args.to_dict(), {'--knob': True, '1': False})

  def test_to_list_bool(self):
    args = job_blocks.SequentialArgs.from_collection({'yes': True, 'no': False})

    self.assertEqual(args.to_list(str), ['--yes', '--nono'])

  def test_to_list_none(self):
    args = job_blocks.SequentialArgs.from_collection(
        {'skip_me': None, 'pass_me': 'None'}
    )

    self.assertEqual(args.to_list(str), ['--pass_me=None'])

  def test_sequential_args_from_string(self):
    with self.assertRaisesRegex(
        ValueError,
        (
            "Tried to construct xm.SequentialArgs from a string: '--foo'. "
            "Wrap it in a list: \\['--foo'\\] to make it a single argument."
        ),
    ):
      job_blocks.SequentialArgs.from_collection('--foo')

  def test_get_args_for_all_jobs(self):
    group = job_blocks.JobGroup(
        a=job_blocks.Job(mock.Mock(), mock.Mock()),
        b=job_blocks.JobGroup(
            b1=job_blocks.JobGroup(
                b1i=job_blocks.Job(mock.Mock(), mock.Mock()),
                b1ii=job_blocks.Job(mock.Mock(), mock.Mock()),
            ),
            b2=job_blocks.Job(mock.Mock(), mock.Mock()),
        ),
    )
    logdir = {'logdir': '/logdir/1'}
    expected = {
        'a': {'args': logdir},
        'b': {
            'b1': {
                'b1i': {'args': logdir},
                'b1ii': {'args': logdir},
            },
            'b2': {'args': logdir},
        },
    }
    args = job_blocks.get_args_for_all_jobs(group, logdir)
    self.assertDictEqual(expected, args)


if __name__ == '__main__':
  absltest.main()



# type: ignore

import IPython
from absl import app
from traitlets.config import Config

from lxm3.clusters import gridengine
from lxm3.xm_cluster import config as config_lib
from lxm3.xm_cluster.console import console


def main(_):
    config = config_lib.default()
    cluster_settings = config.cluster_settings()

    console.log(
        "Creating a client to cluster "
        f"{cluster_settings.user}@{cluster_settings.hostname} ..."
    )

    c = Config()
    c.InteractiveShellApp.exec_lines = [
        "%load_ext rich",
    ]

    client = gridengine.GridEngineCluster(
        cluster_settings.hostname, cluster_settings.user
    )
    IPython.start_ipython(argv=[], config=c, user_ns={"client": client})


if __name__ == "__main__":
    app.run(main)



from lxm3 import xm
from lxm3.xm_cluster import artifacts
from lxm3.xm_cluster import config as config_lib
from lxm3.xm_cluster.packaging import cluster


def package_for_local_executor(packageable: xm.Packageable):
    config = config_lib.default()
    local_settings = config_lib.default().local_settings()
    artifact_store = artifacts.LocalArtifactStore(
        local_settings.storage_root, project=config.project()
    )
    return cluster._PACKAGING_ROUTER(
        packageable.executable_spec, packageable, artifact_store
    )



import os
import shutil
import subprocess
import sys
import unittest
import zipfile
from unittest import mock
from unittest.mock import patch

from absl.testing import absltest
from absl.testing import parameterized

from lxm3 import xm
from lxm3 import xm_cluster
from lxm3.clusters import gridengine as gridengine_cluster
from lxm3.clusters import slurm as slurm_cluster
from lxm3.xm_cluster import JobRequirements
from lxm3.xm_cluster import artifacts
from lxm3.xm_cluster import config
from lxm3.xm_cluster import executables
from lxm3.xm_cluster import executors
from lxm3.xm_cluster.execution import gridengine
from lxm3.xm_cluster.execution import job_script
from lxm3.xm_cluster.execution import local
from lxm3.xm_cluster.execution import slurm


def is_singularity_installed():
    return shutil.which("singularity") is not None


def is_docker_installed():
    return shutil.which("docker") is not None


class JobScriptBuilderTest(parameterized.TestCase):
    def test_env_vars(self):
        env_var_str = job_script._create_env_vars([{"FOO": "BAR1"}, {"FOO": "BAR2"}])
        expected = """\
FOO_0="BAR1"
FOO_1="BAR2"
FOO=$(eval echo \\$"FOO_$LXM_TASK_ID")
export FOO"""
        self.assertEqual(env_var_str, expected)

    def test_empty_env_vars(self):
        self.assertEqual(job_script._create_env_vars([{}]), "")

    def test_different_keys(self):
        with self.assertRaises(ValueError):
            job_script._create_env_vars([{"FOO": "BAR1"}, {"BAR": "BAR2"}])

    def test_args(self):
        args_str = job_script._create_args(
            [["--seed=1", "--task=1"], ["--seed=2", "--task=2"]]
        )
        expected = """\
TASK_CMD_ARGS_0="--seed=1 --task=1"
TASK_CMD_ARGS_1="--seed=2 --task=2"
TASK_CMD_ARGS=$(eval echo \\$"TASK_CMD_ARGS_$LXM_TASK_ID")
echo $TASK_CMD_ARGS"""
        self.assertEqual(args_str, expected)

    def test_empty_args(self):
        self.assertEqual(job_script._create_args([]), ":;")
        self.assertEqual(
            job_script._create_args([[]]),
            """\
TASK_CMD_ARGS_0=""
TASK_CMD_ARGS=$(eval echo \\$"TASK_CMD_ARGS_$LXM_TASK_ID")
echo $TASK_CMD_ARGS""",
        )

    def test_get_additional_env(self):
        job_env = {"FOO": "FOO_0", "OVERRIDE": "OVERRIDE"}

        self.assertEqual(
            job_script._get_additional_env(job_env, {"FOO": "FOO_HOST", "BAR": "BAR"}),
            {"BAR": "BAR"},
        )

    def test_additional_bindings(self):
        job_binds = {"/c": "/b", "/d": "/e"}
        overrides = {"/a": "/b", "/foo": "/bar"}
        additional_binds = job_script._get_additional_binds(job_binds, overrides)
        self.assertEqual(additional_binds, {"/foo": "/bar"})


class LocalExecutionTest(parameterized.TestCase):
    @parameterized.named_parameters(
        ("cpu", None, False),
        ("gpu", "/usr/bin/nvidia-smi", True),
    )
    def test_local_infer_gpu_request(self, nvidia_smi_path, expected):
        executor = executors.Local()
        with patch("shutil.which", return_value=nvidia_smi_path):
            use_gpu = local.LocalJobScriptBuilder._is_gpu_requested(executor)
            self.assertEqual(use_gpu, expected)

    def test_local_launch(self):
        staging_dir = self.create_tempdir(name="staging")

        archive_name = "archive.zip"
        container_name = "container.sif"
        archive = staging_dir.create_file(archive_name)
        container = staging_dir.create_file(container_name)

        deploy_dir = self.create_tempdir(name="deploy")
        executable = executables.Command(
            name="test",
            entrypoint_command="echo hello",
            resource_uri=archive.full_path,
            singularity_image=container.full_path,
        )
        executor = executors.Local()
        job = xm.Job(executable, executor, name="test")
        artifact = artifacts.LocalArtifactStore(deploy_dir.full_path)
        settings = config.LocalSettings()
        client = local.LocalClient(settings, artifact)
        with mock.patch.object(subprocess, "run"):
            client.launch("test_job", job)

    def _run_job_script(self, job_script_content, env=None):
        job_script = self.create_tempfile("job.sh", content=job_script_content)
        os.chmod(job_script.full_path, 0o755)
        workdir = self.create_tempdir("workdir").full_path

        try:
            process = subprocess.run(
                [job_script.full_path],
                check=True,
                env=env,
                capture_output=True,
                cwd=workdir,
            )
        except subprocess.CalledProcessError as e:
            print("job_script:", job_script_content)
            print("stdout:", e.stdout.decode("utf-8"))
            print("stderr:", e.stderr.decode("utf-8"))
            raise
        return process

    def test_job_script_run_single_job(self):
        tmpf = self.create_tempfile("test.zip")
        with zipfile.ZipFile(tmpf.full_path, "w") as z:
            info = zipfile.ZipInfo("entrypoint.sh")
            info.external_attr = 0o777 << 16  # give full access to included file
            z.writestr(
                info,
                """\
#!/usr/bin/env bash
echo $@ $FOO""",
            )
        executable = xm_cluster.Command(
            "foo", "./entrypoint.sh", resource_uri=tmpf.full_path
        )
        job = xm.Job(
            executable,
            executor=xm_cluster.Local(),
            args={"seed": 1},
            env_vars={"FOO": "FOO_0"},
        )
        builder = local.LocalJobScriptBuilder()
        job_script_content = builder.build(job, "foo", "/tmp")
        process = self._run_job_script(job_script_content)
        self.assertEqual(process.stdout.decode("utf-8").strip(), "--seed=1 FOO_0")

    def test_job_script_run_array_job(self):
        tmpf = self.create_tempfile("test.zip")
        with zipfile.ZipFile(tmpf.full_path, "w") as z:
            info = zipfile.ZipInfo("entrypoint.sh")
            info.external_attr = 0o777 << 16  # give full access to included file
            z.writestr(
                info,
                """\
#!/usr/bin/env bash
echo $@ $FOO""",
            )
        executable = xm_cluster.Command(
            "foo", "./entrypoint.sh", resource_uri=tmpf.full_path
        )
        job = xm_cluster.ArrayJob(
            executable,
            executor=xm_cluster.Local(),
            args=[{"seed": 1}, {"seed": 2}],
            env_vars=[{"FOO": "FOO_0"}, {"FOO": "FOO_1"}],
        )
        builder = local.LocalJobScriptBuilder()
        job_script_content = builder.build(job, "foo", "/tmp")
        process = self._run_job_script(
            job_script_content, env={builder.TASK_ID_VAR_NAME: "1"}
        )
        self.assertEqual(process.stdout.decode("utf-8").strip(), "--seed=1 FOO_0")
        process = self._run_job_script(
            job_script_content, env={builder.TASK_ID_VAR_NAME: "2"}
        )
        self.assertEqual(process.stdout.decode("utf-8").strip(), "--seed=2 FOO_1")

    def test_job_script_handles_ml_collections_quoting(self):
        tmpf = self.create_tempfile("test.zip")
        with zipfile.ZipFile(tmpf.full_path, "w") as z:
            info = zipfile.ZipInfo("entrypoint.sh")
            info.external_attr = 0o777 << 16  # give full access to included file
            z.writestr(
                "run.py",
                """\
from absl import app
from ml_collections import config_dict
from ml_collections import config_flags

def _get_config():
    config = config_dict.ConfigDict()
    config.name = ""
    return config


_CONFIG = config_flags.DEFINE_config_dict("config", _get_config())


def main(_):
    config = _CONFIG.value
    print(config.name)

if __name__ == "__main__":
    config = _get_config()
    app.run(main)

""",
            )
            z.writestr(
                info,
                """\
#!/usr/bin/env bash
{} run.py $@
""".format(
                    sys.executable
                ),
            )
        executable = xm_cluster.Command(
            "foo", "./entrypoint.sh", resource_uri=tmpf.full_path
        )
        job = xm_cluster.ArrayJob(
            executable,
            executor=xm_cluster.Local(),
            args=[{"config.name": "train[:90%]"}],
            env_vars=[{"FOO": "FOO_0"}],
        )
        builder = local.LocalJobScriptBuilder()
        job_script_content = builder.build(job, "foo", "/tmp")
        process = self._run_job_script(job_script_content, env={"SGE_TASK_ID": "1"})
        self.assertEqual(process.stdout.decode("utf-8").strip(), "train[:90%]")

    @unittest.skipIf(not is_singularity_installed(), "Singularity is not installed")
    def test_singularity(self):
        tmpf = self.create_tempfile("test.zip")
        with zipfile.ZipFile(tmpf.full_path, "w") as z:
            info = zipfile.ZipInfo("entrypoint.sh")
            info.external_attr = 0o777 << 16  # give full access to included file
            z.writestr(
                info,
                """\
#!/usr/bin/env bash
echo $FOO""",
            )

        executable = xm_cluster.Command(
            "test",
            "./entrypoint.sh",
            resource_uri=tmpf.full_path,
            singularity_image="docker://python:3.10-slim",
        )
        job = xm_cluster.ArrayJob(
            executable, executor=xm_cluster.Local(), env_vars=[{"FOO": "FOO_0"}]
        )
        job_script_content = local.LocalJobScriptBuilder().build(job, "foo", "/tmp")
        process = self._run_job_script(
            job_script_content, env={local.LocalJobScriptBuilder.TASK_ID_VAR_NAME: "1"}
        )
        self.assertEqual(process.stdout.decode("utf-8").strip(), "FOO_0")

    @unittest.skipIf(not is_docker_installed(), "Docker is not installed")
    def test_docker_image(self):
        tmpf = self.create_tempfile("test.zip")
        with zipfile.ZipFile(tmpf.full_path, "w") as z:
            info = zipfile.ZipInfo("entrypoint.sh")
            info.external_attr = 0o777 << 16  # give full access to included file
            z.writestr(
                info,
                """\
#!/usr/bin/env bash
echo $FOO""",
            )

        executable = xm_cluster.Command(
            "test",
            "./entrypoint.sh",
            resource_uri=tmpf.full_path,
            docker_image="python:3.10-slim",
        )
        job = xm_cluster.ArrayJob(
            executable, executor=xm_cluster.Local(), env_vars=[{"FOO": "FOO_0"}]
        )
        job_script_content = local.LocalJobScriptBuilder().build(job, "foo", "/tmp")
        process = self._run_job_script(
            job_script_content, env={local.LocalJobScriptBuilder.TASK_ID_VAR_NAME: "1"}
        )
        self.assertEqual(process.stdout.decode("utf-8").strip(), "FOO_0")


class GridEngineExecutionTest(parameterized.TestCase):
    @parameterized.parameters(
        (executors.GridEngine(requirements=JobRequirements(gpu=1)), True),
        (executors.GridEngine(parallel_environments={"gpu": 1}), True),
        (executors.GridEngine(), False),
    )
    def test_sge_infer_gpu_request(self, executor, expected):
        use_gpu = gridengine.GridEngineJobScriptBuilder._is_gpu_requested(executor)
        self.assertEqual(use_gpu, expected)

    def test_gridengine_header(self):
        executor = executors.GridEngine(
            resources={"h_vmem": "1G"},
            parallel_environments={"gpu": 1},
            project="test",
            account="alloc",
            walltime=10 * xm.Min,
            extra_directives=["-test_extra 1"],
            queue="test_queue",
        )
        header = gridengine.header_from_executor("test_job", executor, None, "/logs")
        self.assertIn("#$ -l h_vmem=1G", header)
        self.assertIn("#$ -l h_rt=00:10:00", header)
        self.assertIn("#$ -pe gpu 1", header)
        self.assertIn("#$ -P test", header)
        self.assertIn("#$ -A alloc", header)
        self.assertIn("#$ -q test_queue", header)
        self.assertIn("#$ -test_extra 1", header)

        executor = executors.GridEngine(max_parallel_tasks=2)
        header = gridengine.header_from_executor("test_job", executor, 10, "/logs")
        self.assertIn("#$ -t 1-10", header)
        self.assertIn("#$ -tc 2", header)

    def test_setup_cmds(self):
        executable = executables.Command(
            name="test",
            entrypoint_command="echo hello",
            resource_uri="",
            singularity_image="docker://python:3.10-slim",
        )
        executor = executors.GridEngine(modules=["module1"])
        with patch.object(
            gridengine.GridEngineJobScriptBuilder,
            "_is_gpu_requested",
            return_value=True,
        ):
            setup_cmds = gridengine.GridEngineJobScriptBuilder._create_setup_cmds(
                executable, executor
            )
        self.assertIn("nvidia-smi", setup_cmds)
        self.assertIn("module load module1", setup_cmds)
        self.assertIn("singularity --version", setup_cmds)

    def test_launch(self):
        staging_dir = self.create_tempdir(name="staging")

        archive_name = "archive.zip"
        container_name = "container.sif"
        archive = staging_dir.create_file(archive_name)
        container = staging_dir.create_file(container_name)

        deploy_dir = self.create_tempdir(name="deploy")
        executable = executables.Command(
            name="test",
            entrypoint_command="echo hello",
            resource_uri=archive.full_path,
            singularity_image=container.full_path,
        )
        executor = executors.GridEngine()
        job = xm.Job(executable, executor, name="test")
        artifact = artifacts.LocalArtifactStore(deploy_dir.full_path)
        settings = config.ClusterSettings()
        client = gridengine.GridEngineClient(settings, artifact)
        with mock.patch.object(
            gridengine_cluster.GridEngineCluster, "launch"
        ) as mock_launch:
            mock_launch.return_value = gridengine_cluster.parse_job_id("1")
            client.launch("test_job", job)


class SlurmExecutionTest(parameterized.TestCase):
    def test_slurm_header(self):
        executor = executors.Slurm(
            resources={"mem": "1G"},
            walltime=10 * xm.Hr,
            exclusive=True,
            partition="contrib-gpu-long",
            extra_directives=["--test_extra 1"],
        )
        header = slurm.header_from_executor("test_job", executor, None, "/logs")
        self.assertIn("#SBATCH --mem=1G", header)
        self.assertIn("#SBATCH --time=10:00:00", header)
        self.assertIn("#SBATCH --exclusive", header)
        self.assertIn("#SBATCH --partition=contrib-gpu-long", header)
        self.assertIn("#SBATCH --test_extra 1", header)

        executor = executors.Slurm()
        header = slurm.header_from_executor("test_job", executor, 10, "/logs")
        self.assertIn("#SBATCH --array=1-10", header)

    def test_setup_cmds(self):
        executable = executables.Command(
            name="test",
            entrypoint_command="echo hello",
            resource_uri="",
            singularity_image="docker://python:3.10-slim",
        )
        executor = executors.Slurm(modules=["module1"])
        with patch.object(
            gridengine.GridEngineJobScriptBuilder,
            "_is_gpu_requested",
            return_value=True,
        ):
            setup_cmds = slurm.SlurmJobScriptBuilder._create_setup_cmds(
                executable, executor
            )
        self.assertIn("module load module1", setup_cmds)
        self.assertIn("singularity --version", setup_cmds)

    def test_slurm_launch(self):
        staging_dir = self.create_tempdir(name="staging")

        archive_name = "archive.zip"
        container_name = "container.sif"
        archive = staging_dir.create_file(archive_name)
        container = staging_dir.create_file(container_name)

        deploy_dir = self.create_tempdir(name="deploy")
        executable = executables.Command(
            name="test",
            entrypoint_command="echo hello",
            resource_uri=archive.full_path,
            singularity_image=container.full_path,
        )
        executor = executors.Slurm()
        job = xm.Job(executable, executor, name="test")
        artifact = artifacts.LocalArtifactStore(deploy_dir.full_path)
        settings = config.ClusterSettings()
        client = slurm.SlurmClient(settings, artifact)
        with mock.patch.object(slurm_cluster.SlurmCluster, "launch") as mock_launch:
            mock_launch.return_value = slurm_cluster.parse_job_id("job 1")
            client.launch("test_job", job)


if __name__ == "__main__":
    absltest.main()



import datetime

from absl.testing import absltest
from absl.testing import parameterized

from lxm3.xm_cluster import executors
from lxm3.xm_cluster.execution import gridengine
from lxm3.xm_cluster.execution import slurm


class ExecutorTest(parameterized.TestCase):
    @parameterized.parameters(
        (None, None),
        (5 * 3600 + 20 * 60 + 25, datetime.timedelta(hours=5, minutes=20, seconds=25)),
        (datetime.timedelta(hours=3), datetime.timedelta(hours=3)),
    )
    def test_convert_time(self, input, expected):
        self.assertEqual(executors._convert_time(input), expected)

    @parameterized.parameters(
        (datetime.timedelta(hours=5, minutes=20, seconds=25), "05:20:25"),
        (datetime.timedelta(hours=55, minutes=20, seconds=25), "55:20:25"),
        (datetime.timedelta(hours=102, minutes=20, seconds=25), "102:20:25"),
    )
    def test_format_sge_time(self, input, expected):
        self.assertEqual(gridengine._format_time(input.total_seconds()), expected)

    @parameterized.parameters(
        (datetime.timedelta(hours=5, minutes=20, seconds=25), "05:20:25"),
        (datetime.timedelta(hours=55, minutes=20, seconds=25), "02-07:20:25"),
    )
    def test_format_slurm_time(self, input, expected):
        self.assertEqual(slurm._format_slurm_time(input), expected)


if __name__ == "__main__":
    absltest.main()



import os
import unittest.mock
import zipfile

from absl.testing import absltest
from absl.testing import parameterized

from lxm3 import xm
from lxm3 import xm_cluster
from lxm3.xm_cluster import artifacts
from lxm3.xm_cluster.packaging import cluster

_HERE = os.path.abspath(os.path.dirname(__file__))


class PackagingTest(parameterized.TestCase):
    @parameterized.parameters(
        (cluster._package_python_package,),
    )
    def test_package_python(self, pkg_fun):
        spec = xm_cluster.PythonPackage(
            entrypoint=xm_cluster.ModuleName("py_package.main"),
            path=os.path.join(_HERE, "testdata/test_pkg"),
        )

        with unittest.mock.patch("subprocess.run"):
            tmpdir = self.create_tempdir().full_path
            store = artifacts.LocalArtifactStore(tmpdir, "test")
            executable = pkg_fun(
                spec,
                xm.Packageable(
                    spec,
                    xm_cluster.Local().Spec(),
                ),
                store,
            )
            self.assertIsInstance(executable, xm_cluster.Command)

    def test_package_default_pip_args(self):
        spec = xm_cluster.PythonPackage(
            entrypoint=xm_cluster.ModuleName("py_package.main"),
            path=os.path.join(_HERE, "testdata/test_pkg"),
        )
        self.assertEqual(
            sorted(spec.pip_args),
            sorted(["--no-deps", "--no-compile"]),
        )

    @parameterized.parameters(
        (cluster._package_universal_package,),
    )
    def test_package_universal(self, pkg_fun):
        spec = xm_cluster.UniversalPackage(
            entrypoint=["python3", "main.py"],
            path=os.path.join(_HERE, "testdata/test_universal"),
            build_script="build.sh",
        )
        tmpdir = self.create_tempdir().full_path
        store = artifacts.LocalArtifactStore(tmpdir, "test")
        executable = pkg_fun(
            spec,
            xm.Packageable(
                spec,
                xm_cluster.Local().Spec(),
            ),
            store,
        )
        self.assertIsInstance(executable, xm_cluster.Command)
        # Check that archive exists
        self.assertTrue(store._fs.exists(executable.resource_uri))
        with store._fs.open(executable.resource_uri, "rb") as f:
            archive = zipfile.ZipFile(f)  # type: ignore
            self.assertEqual(set(archive.namelist()), set(["main.py"]))

    def test_package_universal_not_executable(self):
        with self.assertRaises(ValueError):
            xm_cluster.UniversalPackage(
                entrypoint=["python3", "main.py"],
                path=os.path.join(_HERE, "testdata/test_universal"),
                build_script="build_not_executable.sh",
            )

    def test_package_singularity_invalid_path(self):
        py_package = xm_cluster.PythonPackage(
            entrypoint=xm_cluster.ModuleName("py_package.main"),
            path=os.path.join(_HERE, "testdata/test_pkg"),
        )
        fake_image_path = self.create_tempfile().full_path
        # This is OK
        xm_cluster.SingularityContainer(py_package, image_path=fake_image_path)
        # Raises on non-existent path
        with self.assertRaises(ValueError):
            xm_cluster.SingularityContainer(
                py_package, image_path=os.path.join(_HERE, "/fake/image.sif")
            )


if __name__ == "__main__":
    absltest.main()



import os
import pathlib
from unittest import mock

import fabric
from absl.testing import absltest
from absl.testing import parameterized

from lxm3.clusters import gridengine


class GridEngineTest(parameterized.TestCase):
    @parameterized.named_parameters(
        [
            {
                "testcase_name": "array job",
                "text": 'Your job-array 9830192.1-5:1 ("MyTESTJOBNAME") has been submitted',  # noqa
                "expected": [
                    "9830192.1",
                    "9830192.2",
                    "9830192.3",
                    "9830192.4",
                    "9830192.5",
                ],
            },
            {
                "testcase_name": "job",
                "text": 'Your job 9830196 ("MyTESTJOBNAME") has been submitted',
                "expected": ["9830196"],
            },
            {
                "testcase_name": "job name with numbers",
                "text": 'Your job-array 9834008 ("job-1677559794797") has been submitted',  # noqa
                "expected": ["9834008"],
            },
            {
                "testcase_name": "array-job name with numbers",
                "text": 'Your job-array 9834008.1-5:1 ("job-1677559794797") has been submitted',  # noqa
                "expected": [f"9834008.{i}" for i in range(1, 6)],
            },
        ]
    )
    def test_parse_job_id(self, text, expected):
        match = gridengine.parse_job_id(text)
        self.assertEqual(gridengine.split_job_ids(match), expected)

    def test_parse_invalid_job_id(self):
        with self.assertRaises(ValueError):
            gridengine.parse_job_id("Failed")

    def test_parse_job_2(self):
        text = 'Your job-array 9834008.1-5:1 ("job-1677559794797") has been submitted'
        match = gridengine.parse_job_id(text)
        self.assertEqual(match.group(0), "9834008.1-5:1")

    @parameterized.named_parameters(
        [
            {
                "testcase_name": "qstat1",
                "input": "qstat.xml",
                "expected": {"9830166": {"state": "pending"}},
            },
            {
                "testcase_name": "qstat2",
                "input": "qstat2.xml",
                "expected": {
                    "9830188.1": {"state": "pending"},
                    "9830188.2": {"state": "pending"},
                    "9830188.3": {"state": "pending"},
                    "9830188.4": {"state": "pending"},
                    "9830188.5": {"state": "pending"},
                },
            },
            {
                "testcase_name": "qstat3",
                "input": "qstat3.xml",
                "expected": {
                    "9830197.1": {"state": "running"},
                    "9830197.2": {"state": "running"},
                    "9830197.3": {"state": "running"},
                    "9830197.4": {"state": "running"},
                    "9830197.5": {"state": "running"},
                },
            },
            {
                "testcase_name": "qstat4",
                "input": "qstat4.xml",
                "expected": {
                    "9830420.1": {"state": "running"},
                    "9830420.2": {"state": "running"},
                    "9830421.1": {"state": "pending"},
                    "9830421.2": {"state": "pending"},
                },
            },
        ]
    )
    def test_parse_qstat_state(self, input, expected):
        xml_output = pathlib.Path(
            os.path.join(os.path.dirname(__file__), "testdata", input)
        ).read_text()
        infos = gridengine.parse_qstat(xml_output)
        self.assertEqual(infos, expected)

    @parameterized.parameters(
        ([12, 13, 14, 16], ["12-14", "16"]),
        (reversed([12, 13, 14, 16]), ["12-14", "16"]),
        ([1, 3, 14, 15, 16], ["1", "3", "14-16"]),
        (["1", " 3", 14, 15, 16], ["1", "3", "14-16"]),
    )
    def test_merge_taskids(self, input, expected):
        actual = gridengine.merge_taskids(input)
        self.assertEqual(actual, expected)


class AccountingTest(absltest.TestCase):
    def test_accounting(self):
        DATA1 = (
            pathlib.Path(__file__).parent.joinpath("testdata/qacct1.txt").read_text()
        )
        DATA2 = (
            pathlib.Path(__file__).parent.joinpath("testdata/qacct2.txt").read_text()
        )
        self.assertLen(gridengine.parse_accounting(DATA1), 2)
        self.assertLen(gridengine.parse_accounting(DATA2), 2)


class ClusterTest(absltest.TestCase):
    @mock.patch("fabric.Connection")
    def test_cluster(self, mock_connection):
        instance = mock_connection.return_value
        instance.run.return_value = fabric.Result(
            connection=instance,
            stdout='Your job 9830196 ("MyTESTJOBNAME") has been submitted',
        )
        cluster = gridengine.GridEngineCluster(hostname="host", username="user")
        match = cluster.launch("job.qsub")
        self.assertEqual(match.group(0), "9830196")
        cluster.close()


if __name__ == "__main__":
    absltest.main()



from typing import Dict, Optional

import attr

from lxm3 import xm


@attr.s(auto_attribs=True)
class Command(xm.Executable):
    entrypoint_command: str
    resource_uri: str
    args: xm.SequentialArgs = attr.Factory(xm.SequentialArgs)
    env_vars: Dict[str, str] = attr.Factory(dict)
    singularity_image: Optional[str] = None
    docker_image: Optional[str] = None
